import os
import torch
import helpers.ascon_helper as ascon

sbox_chunk_size = 8
parameters = {
    "dataset_name"        : '../data/ASCON_Encrypt_60000_20230404_132906_all_mlpsca_ready.trs',#Random key
    "dataset_size"        : 60000,
    "training_size"       : 50000,
    "val_size"            : 10000,
    "subkey"              : 1,
    "num_key_hypotheses"  : 2**sbox_chunk_size,
    "n_samples"           : 772,#Random key
    "r"                   : 0.8,
    "n_repeat"            : 50,
    "ge_rate"             : 5,
    "shuffle"             : True,
    "chunk_size"          : sbox_chunk_size,
}

model = {
    "loss_function"       : torch.nn.CrossEntropyLoss(),
    "lr"                  : 1e-05,
    "bs"                  : 128,
    "num_epochs"          : 500,
}

callbacks = {
    "modelcheckpoint":{
        "period":-1,
    },
    "guessingentropy":{
        "ge_rate":5,
        "n_repeat":5,
        "r":0.6,
    },
    "tensorboard":{
        "training_accuracy":True,
        "validation_accuracy":True,
        "training_loss":True,
        "validation_loss":True,
    },
}

def get_attack_parameters(traceset, subkey, index):     
    return {'nonce0':traceset["nonce0"][index], 
            'nonce1':traceset["nonce1"][index], 
            'k':traceset["key"][index],
            'subkey':subkey}

def get_known_key(traceset, subkey, index):
    return ascon.select_subkey_sbox_chunk(subkey, traceset['key'][index], parameters["chunk_size"])

def intermediate_value(keyguess, trace_parameters):
    return ascon.intermediate_value_sbox_chunk(keyguess,trace_parameters, parameters["chunk_size"])


# eSCAPyDL - a Deep Learning Side-Channel Analysis Python Framework
# Copyright (C) 2023  Weissbart Léo

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging 
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
logger.debug(device)

import sys
import time
import numpy as np
import collections
import Callbacks
import Utils


#---------------------------------------------------#
# Models

class MLP(nn.Module):
    default_Hparams = {"activation_function":'tanh',
                        "lr": 1e-3,
                        "n_layers":2,
                        "n_units_l0": 20, 
                        "n_units_l1": 10}
    def __init__(self, parameters, Hparams={}):
        super(MLP, self).__init__()
        self.name = "MLP"
        self.Hparams = {**MLP.default_Hparams, **Hparams}#Merge default and given Hparams
        self.n_layers = int(self.Hparams['n_layers'])
        self.act = {"relu":nn.ReLU(), "tanh":nn.Tanh(), "selu":nn.SELU()}[self.Hparams['activation_function'].lower()] if 'activation_function' in self.Hparams.keys() else nn.ReLU()
        self.hidden = nn.ModuleList()
        self.hidden.append(nn.Linear(parameters["n_samples"], int(self.Hparams['n_units_l0'])))
        for i in range(self.n_layers-1):
            self.hidden.append(nn.Linear(int(self.Hparams['n_units_l{}'.format(i)]), int(self.Hparams['n_units_l{}'.format(i+1)])))
        self.fc1 = nn.Linear(int(self.Hparams['n_units_l{}'.format(self.n_layers-1)]), parameters["num_classes"])
        self.dropout = nn.Dropout(0.1)
        self.bnorm = nn.BatchNorm1d(parameters["n_samples"])
        # self.bnorm0 = nn.BatchNorm1d(1)
        self.flatten = nn.Flatten()
    def forward(self, x):
        # x = self.bnorm0(x)
        x = self.flatten(x)
        x = self.bnorm(x)
        for h_layer in self.hidden:
            x = self.act(h_layer(x))
            x = self.dropout(x)
        x = self.fc1(x)
        return x

    def fit(self,*args, **kwargs):
        if "Hparams" in kwargs.keys():
            kwargs["Hparams"] = {**kwargs["Hparams"], **self.Hparams}
        else:
            kwargs["Hparams"] = self.Hparams
        return fit(self, *args, **kwargs)

class CNN_onetrace(nn.Module):
    def __init__(self, parameters, Hparams={}):
        super(CNN_onetrace, self).__init__()
        logger.debug("Create CNN module.")
        logger.debug(">> Input size:{}, num_classes:{}".format(parameters["n_samples"], parameters["num_classes"]))
        self.name = "CNN"
        self.n_samples = parameters["n_samples"]
        self.num_blocks = len(bin(self.n_samples)[2:]) - 2
        self.cnn_output_size = self.n_samples >> self.num_blocks
        logger.debug(">> Number of hidden conv blocks:{}, cnn output size:{}".format(self.num_blocks, self.cnn_output_size>>1))
        self.cnn_sequence = nn.Sequential(collections.OrderedDict([
            ('batchnorm0', nn.BatchNorm1d(1)),
            ('conv1', nn.Conv1d(1, 8, kernel_size=3, padding='same')),
            ('batchnorm1', nn.BatchNorm1d(8)),
            ('maxpool1', nn.MaxPool1d(2))
            ])
        )
        for i in range(0,self.num_blocks):
            self.cnn_sequence.append(nn.Conv1d(min(512,8<<i), min(512,8<<(i+1)),kernel_size=3, padding='same'))
            self.cnn_sequence.append(nn.ReLU())
            if i%2==1:
                self.cnn_sequence.append(nn.BatchNorm1d(min(512,8<<(i+1))))
            self.cnn_sequence.append(nn.MaxPool1d(2))
        self.cnn_sequence.append(nn.Flatten())
        self.cnn_sequence.append(nn.Dropout(0.5))
        #input size is the number of filters in last conv layer * output size of last conv layer
        self.cnn_sequence.append(nn.Linear(min(512,8<<self.num_blocks)*(self.cnn_output_size>>1) ,512))
        self.cnn_sequence.append(nn.ReLU())
        self.cnn_sequence.append(nn.Dropout(0.5))
        self.cnn_sequence.append(nn.Linear(512,parameters["num_classes"]))

    def forward(self, x):
        x = self.cnn_sequence(x)
        return x
    
    def fit(self,*args, **kwargs):
        return fit(self, *args, **kwargs)
    
class CNN(nn.Module):
    default_Hparams = {"activation_function": "tanh", 
                           "pooling": "average",
                           "optimizer": "adam",
                           "weight_init": "he_uniform",
                           "n_conv_block": 2,
                           "kernel_size_0": 11, 
                           "c_out_0": 3,
                           "kernel_size_1": 6,
                           "c_out_1": 3, 
                           "lr": 1e-3,  
                           "n_layers": 2, 
                           "n_units_l0": 20, 
                           "n_units_l1": 10}
    def __init__(self, parameters, Hparams={}, *args, **kwargs) -> None:
        super(CNN, self).__init__(*args, **kwargs)
        self.name = "CNN"
        self.Hparams = {**CNN.default_Hparams, **Hparams}#Merge default and given Hparams
        self.n_conv_blocks = int(self.Hparams["n_conv_block"])
        self.n_fc_layers = int(self.Hparams["n_layers"])
        self.act = {"relu":nn.ReLU(), "tanh":nn.Tanh(), "selu":nn.SELU()}[self.Hparams['activation_function'].lower()] if 'activation_function' in self.Hparams.keys() else nn.ReLU()
        #default conv1d parameters
        padding = int(self.Hparams["padding"]) if "padding" in self.Hparams.keys() else 0
        stride = int(self.Hparams["stride"]) if "stride" in self.Hparams.keys() else 1
        dilation = int(self.Hparams["dilation"]) if "dilatation" in self.Hparams.keys() else 1
        pooling_ks = int(self.Hparams["pooling_ks"]) if "pooling_ks" in self.Hparams.keys() else 2
        pooling_stride = int(self.Hparams["pooling_stride"]) if "pooling_stride" in self.Hparams.keys() else pooling_ks
        pooling_padding = int(self.Hparams["pooling_padding"]) if "pooling_padding" in self.Hparams.keys() else 0
        self.pooling = {"none":None, "max":nn.MaxPool1d, "average":nn.AvgPool1d}[self.Hparams['pooling'].lower()] if 'pooling' in self.Hparams.keys() else None
        if self.pooling is not None:
            self.pooling = self.pooling(pooling_ks, stride=pooling_stride, padding=pooling_padding)
        
        i=0
        in_features = parameters["n_samples"]
        c_in = 1
        self.cnn_layers = nn.ModuleList()
        self.input_bn = nn.BatchNorm1d(c_in)
        for i in range(self.n_conv_blocks):
            c_out = int(self.Hparams["c_out_{}".format(i)])
            self.cnn_layers.append(nn.Conv1d(c_in, c_out, kernel_size=int(self.Hparams["kernel_size_{}".format(i)]), padding='same'))
            # out_features = (in_features + 2 * padding - dilation * (int(self.Hparams["kernel_size_{}".format(i)]) - 1) - 1) // stride + 1
            out_features = in_features
            if self.pooling is not None:
                out_features = max(1, (out_features - 2*pooling_padding - pooling_ks) // pooling_stride + 1)
            in_features = out_features
            c_in = c_out
        self.flatten = nn.Flatten()
        in_features = int(c_in * in_features)
        self.fc_input_bn = nn.BatchNorm1d(in_features)
        self.fc_layers = nn.ModuleList()
        for i in range(self.n_fc_layers):
            self.fc_layers.append(nn.Linear(in_features, int(self.Hparams['n_units_l{}'.format(i)])))
            in_features = int(self.Hparams['n_units_l{}'.format(i)])
        self.last_fc = nn.Linear(in_features, parameters["num_classes"])
        self.dropout = nn.Dropout(0.1)
        
    def forward(self, x):
        x = self.input_bn(x)
        for i in range(self.n_conv_blocks):
            x = self.act(self.cnn_layers[i](x))
            if self.pooling is not None:
                x = self.pooling(x)
        x = self.flatten(x)
        x = self.fc_input_bn(x)
        for i in range(self.n_fc_layers):
            x = self.act(self.fc_layers[i](x))
            x = self.dropout(x)
        x = self.last_fc(x)
        return x
    
    def fit(self,*args, **kwargs):
        if "Hparams" in kwargs.keys():
            kwargs["Hparams"] = {**kwargs["Hparams"], **self.Hparams}
        else:
            kwargs["Hparams"] = self.Hparams
        return fit(self, *args, **kwargs)

class payAttentionNN(nn.Module):
    default_Hparams = {"activation_function": "tanh", 
                           "optimizer": "adam",
                           "weight_init": "he_uniform",
                           "reshape_size": 2,
                           "kernel_size": 21,
                           "stride": 10,
                           "units": 128,
                           }
    def __init__(self, parameters, Hparams={}, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.Hparams = {**payAttentionNN.default_Hparams, **Hparams}#Merge default and given Hparams
        self.units = int(self.Hparams['units'])
        self.act = {"relu":nn.ReLU(), "tanh":nn.Tanh(), "selu":nn.SELU()}[self.Hparams['activation_function'].lower()]
        self.reshape_size = int(self.Hparams['reshape_size'])
        self.kernel_size = int(self.Hparams['kernel_size'])
        self.stride = int(self.Hparams['stride'])

        self.local1d = Utils.LocallyConnected1d(1, parameters["n_samples"], 1, self.kernel_size, self.stride, bias=True)
        self.bn1 = nn.BatchNorm1d(self.local1d.output_size//self.reshape_size)
        self.bidir_lstm = nn.LSTM(self.reshape_size,self.units,1,bidirectional=True)

        self.dense_att = nn.Linear(2*self.units, 1, bias=False)
        self.flatten = nn.Flatten()
        self.softmax = nn.Softmax(dim=1)

        self.dropout = nn.Dropout(0.5)
        self.fc1 = nn.Linear(2*self.units, parameters["num_classes"])
        self.bn3 = nn.BatchNorm1d(parameters["num_classes"])

    def forward(self, x):
        #Encoder
        x = self.local1d(x)
        x = torch.reshape(x,(x.shape[0],-1,self.reshape_size))#Cut the sequences into smaller chunks
        x,_ = self.bidir_lstm(x)
        lstm_out = self.bn1(x)
        x = self.act(lstm_out)
        #Attention
        x = self.dense_att(x)
        x = self.flatten(x)
        x = self.softmax(self.bn1(x))
        x = x.repeat(2*self.units,1).view(-1,self.local1d.output_size//self.reshape_size,2*self.units)#Attention probabilities
        #Merge attention and raw lstm output
        x = torch.sum(x*lstm_out, dim=1)
        #Classifier
        x = self.dropout(x)
        x = self.bn3(self.fc1(x))
        return x
    
    def fit(self,*args, **kwargs):
        if "Hparams" in kwargs.keys():
            kwargs["Hparams"] = {**kwargs["Hparams"], **self.Hparams}
        else:
            kwargs["Hparams"] = self.Hparams
        return fit(self, *args, **kwargs)
        
class MLP_byte(nn.Module):
    default_Hparams = {"activation_function":'relu',
                        "n_layers":2,
                        "n_units_l0": 20, 
                        "n_units_l1": 10}
    def __init__(self, parameters, Hparams={}, device_list=[torch.device('cpu')], device_idx=0, *args, **kwargs) -> None:
        super(MLP_byte, self).__init__()
        self.name = "MLP"
        self.device_list = device_list
        self.s_device = device_list[device_idx]
        self.Hparams = {**MLP_byte.default_Hparams, **Hparams}#Merge default and given Hparams
        self.n_layers = int(self.Hparams['n_layers'])
        self.act = {"relu":nn.ReLU(), "tanh":nn.Tanh(), "selu":nn.SELU()}[self.Hparams['activation_function'].lower()] if 'activation_function' in self.Hparams.keys() else nn.ReLU()
        self.hidden = nn.ModuleList()
        self.hidden.append(nn.Linear(parameters["n_samples"], int(self.Hparams['n_units_l0'])))
        for i in range(self.n_layers-1):
            self.hidden.append(nn.Linear(int(self.Hparams['n_units_l{}'.format(i)]), int(self.Hparams['n_units_l{}'.format(i+1)])))
        self.hidden = self.hidden.to(self.s_device)
        self.fc1 = nn.Linear(int(self.Hparams['n_units_l{}'.format(self.n_layers-1)]), parameters["num_classes"]).to(self.s_device)
        self.dropout = nn.Dropout(0.1)
        self.bnorm = nn.BatchNorm1d(parameters["n_samples"]).to(self.s_device)
        self.flatten = nn.Flatten()

    def forward(self, x):
        x = self.flatten(x.to(self.s_device))
        # x = self.bnorm(x)
        for i in range(self.n_layers):
            x = self.act(self.hidden[i](x))
            x = self.dropout(x)
        x = self.fc1(x)
        return x.to(self.device_list[0])

def fit(model, parameters, trainset, valset, Hparams={}, callbacks=[Callbacks.Callback()], verbose=0, *args, **kwargs):
    default_Hparams = {"optimizer":"adam", "weight_init":"he_uniform", "loss_function":"crossentropy", "lr":1e-3, "bs":128, "num_epochs":100}
    Hparams = {**default_Hparams, **Hparams}
    logging.info("fit Hparams: {}".format(Hparams))
    weight_init_func = {"he_uniform":Utils.he_uniform_init_weights, "zero":Utils.zero_init_weights}[Hparams['weight_init'].lower()]
    optimizer_func = {"adam":optim.Adam, "rmsprop":optim.RMSprop, "sgd":optim.SGD}[Hparams['optimizer'].lower()]
    loss_function = {"crossentropy":nn.CrossEntropyLoss(), "mse":nn.MSELoss()}[Hparams['loss_function'].lower()] if isinstance(Hparams['loss_function'], str) else Hparams['loss_function']
    learning_rate = float(Hparams['lr'])
    batch_size = int(Hparams['bs'])
    epochs = int(Hparams['num_epochs'])

    #Create loaders
    trainloader = torch.utils.data.DataLoader(trainset, batch_size=batch_size, shuffle=True, pin_memory=True, num_workers=4)
    valloader = torch.utils.data.DataLoader(valset, batch_size=parameters['val_size'], shuffle=True, pin_memory=True, num_workers=4)#no need for gradient descent

    model.to(device)
    model.apply(weight_init_func)
    optimizer = optimizer_func(model.parameters(), lr=learning_rate, weight_decay=1e-5)
    
    criterion = loss_function
    m = nn.LogSoftmax(dim=1)

    history = {'loss':torch.zeros((epochs), device=device), 'accuracy':torch.zeros((epochs), device=device), 'val_loss':torch.zeros((epochs), device=device), 'val_accuracy':torch.zeros((epochs), device=device)}
    model.train()
    skip = [cb.on_train_begin() for cb in callbacks]
    for epoch in range(epochs):
        start_t = time.time()
        skip = [cb.on_epoch_begin() for cb in callbacks]
        for i, (data, labels) in enumerate(trainloader):
            skip = [cb.on_batch_begin() for cb in callbacks]
            inputs, labels = data.to(device), labels.to(device)
            # forward + backward + optimize
            optimizer.zero_grad()
            outputs = model(inputs)
            loss = criterion(m(outputs), labels.long())
            loss.backward()
            skip = [cb.on_loss_end() for cb in callbacks]
            skip = [cb.on_step_begin() for cb in callbacks]
            optimizer.step()
            skip = [cb.on_step_end() for cb in callbacks]
            history['loss'][epoch] += loss.detach().detach()
            history['accuracy'][epoch] += (outputs.argmax(dim=1) == labels).sum()
        history['loss'][epoch] /= (i+1)
        history['accuracy'][epoch] /= parameters['training_size']

        #Validation pass
        with torch.no_grad():
            for i, (data, labels) in enumerate(valloader):
                inputs, labels = data.to(device), labels.to(device)
                # forward only
                outputs = model(inputs)
                loss = criterion(m(outputs), labels.long())
                history['val_loss'][epoch] += loss
                history['val_accuracy'][epoch] += (outputs.argmax(dim=1) == labels).sum()
            history['val_loss'][epoch] /= (i+1)
            history['val_accuracy'][epoch] /= parameters['val_size']
            skip = [cb.on_epoch_end(model=model, epoch=epoch, history=history, last_epoch=epochs) for cb in callbacks]

        if verbose > 0:
            logger.info("Epoch {}/{} - {:.3f}s - loss: {:.4f} - accuracy: {:.4f} - val_loss: {:.4f} - val_accuracy: {:.4f}".format(epoch+1, epochs, time.time()-start_t, history['loss'][epoch], history['accuracy'][epoch], history['val_loss'][epoch], history['val_accuracy'][epoch]))
    if verbose == 0:
        logger.info("\nEpoch {}/{} - {:.3f}s - loss: {:.4f} - accuracy: {:.4f} - val_loss: {:.4f} - val_accuracy: {:.4f}".format(epoch+1, epochs, time.time()-start_t, history['loss'][epoch], history['accuracy'][epoch], history['val_loss'][epoch], history['val_accuracy'][epoch]))
    training_results = [cb.on_train_end(model) for cb in callbacks]

    #Return all results in history
    np_history = {}
    for d in training_results:
        try:
            np_history = {**np_history, **d}
        except:
            pass
    for key in history.keys():
        np_history[key] = history[key].detach().cpu().numpy()
    return np_history


#---------------------------------------------------#
# Models (multi-target)

class CNN_multi_bytes(nn.Module):
    # TODO: Give a pre-trained CNN on one byte and copy the convlution blocks to freeze it in the network, leaving only the MLPs to train.
    default_Hparams = {"activation_function": "relu", 
                           "pooling": "average",
                           "optimizer": "radam",
                           "weight_init": "he_uniform",
                           "n_conv_block": 2,
                           "kernel_size_0": 30, 
                           "c_out_0": 8,
                           "kernel_size_1": 5,
                           "c_out_1": 16, 
                        #    "kernel_size_2": 6,
                        #    "c_out_2": 32, 
                           "lr": 1e-3,  
                           "n_layers": 4, 
                           "n_units_l0": 500, 
                           "n_units_l1": 470,
                           "n_units_l2": 10,
                           "n_units_l3": 480,
                        #    "n_units_l4": 100,
                        #    "n_units_l5": 100,
                        #    "n_units_l6": 100,
                        #    "n_units_l7": 100,
                           }
    def __init__(self, parameters, Hparams={}, device_list=[torch.device('cuda:{:d}'.format(x)) for x in range(torch.cuda.device_count())], *args, **kwargs) -> None:
        super(CNN_multi_bytes, self).__init__(*args, **kwargs)
        self.name = "multiTargetCNN"
        self.Hparams = {**CNN_multi_bytes.default_Hparams, **Hparams}#Merge default and given Hparams
        self.device_list = device_list if device_list!=[] else [torch.device('cpu')]
        self.nk = len(parameters["subkey"]) if isinstance(parameters["subkey"], list) else 1
        self.nk_per_device = np.ceil(self.nk/len(self.device_list))
        
        self.n_conv_blocks = int(self.Hparams["n_conv_block"])
        self.n_fc_layers = int(self.Hparams["n_layers"])
        self.act = {"relu":nn.ReLU(), "tanh":nn.Tanh(), "selu":nn.SELU()}[self.Hparams['activation_function'].lower()] if 'activation_function' in self.Hparams.keys() else nn.ReLU()
        #default conv1d parameters
        padding = int(self.Hparams["padding"]) if "padding" in self.Hparams.keys() else 0
        stride = int(self.Hparams["stride"]) if "stride" in self.Hparams.keys() else 1
        dilation = int(self.Hparams["dilation"]) if "dilatation" in self.Hparams.keys() else 1
        pooling_ks = int(self.Hparams["pooling_ks"]) if "pooling_ks" in self.Hparams.keys() else 2
        pooling_stride = int(self.Hparams["pooling_stride"]) if "pooling_stride" in self.Hparams.keys() else pooling_ks
        pooling_padding = int(self.Hparams["pooling_padding"]) if "pooling_padding" in self.Hparams.keys() else 0
        self.pooling = {"none":None, "max":nn.MaxPool1d, "average":nn.AvgPool1d}[self.Hparams['pooling'].lower()] if 'pooling' in self.Hparams.keys() else None
        if self.pooling is not None:
            self.pooling = self.pooling(pooling_ks, stride=pooling_stride, padding=pooling_padding).to(self.device_list[0])

        self.dropout = nn.FeatureAlphaDropout(0.1)
        
        in_features = parameters["n_samples"]
        c_in = 1
        self.cnn_layers = nn.ModuleList()
        self.input_bn = nn.BatchNorm1d(c_in).to(self.device_list[0])
        for i in range(self.n_conv_blocks):
            c_out = int(self.Hparams["c_out_{}".format(i)])
            self.cnn_layers.append(nn.Conv1d(c_in, c_out, kernel_size=int(self.Hparams["kernel_size_{}".format(i)])))
            out_features = (in_features + 2 * padding - dilation * (int(self.Hparams["kernel_size_{}".format(i)]) - 1) - 1) // stride + 1
            if self.pooling is not None:
                out_features = max(1, (out_features - 2*pooling_padding - pooling_ks) // pooling_stride + 1)
            in_features = out_features
            c_in = c_out
        self.cnn_layers = self.cnn_layers.to(self.device_list[0])
        in_features = int(c_in * in_features)
        #Divise the branch equally between the number of available devices
        self.multy_byte_mlps = nn.ModuleDict([[str(k), MLP_byte({"n_samples":in_features, "num_classes":parameters["num_classes"]}, device_list=self.device_list, Hparams=self.Hparams, device_idx=int(k//self.nk_per_device))] for k in range(self.nk)])

    def forward(self, x):
        x = self.input_bn(x)
        for i in range(self.n_conv_blocks):
            x = self.act(self.cnn_layers[i](x))
            x = self.dropout(x)
            if self.pooling is not None:
                x = self.pooling(x)
        #Divise the branch equally between the number of available devices
        x = [self.multy_byte_mlps[str(k)](x) for k in range(self.nk)]
        return x
    
    def fit(self,*args, **kwargs):
        if "Hparams" in kwargs.keys():
            kwargs["Hparams"] = {**kwargs["Hparams"], **self.Hparams}
        else:
            kwargs["Hparams"] = self.Hparams
        return multi_fit(self, *args, **kwargs)
    
class MLP_multi_bytes(nn.Module):
    # TODO: Give a pre-trained CNN on one byte and copy the convlution blocks to freeze it in the network, leaving only the MLPs to train.
    default_Hparams = {}
    def __init__(self, parameters, Hparams={}, device_list=[torch.device('cuda:{:d}'.format(x)) for x in range(torch.cuda.device_count())]):
        super(MLP_multi_bytes, self).__init__()
        self.name = "multiTargetMLP"
        self.Hparams = {**MLP_multi_bytes.default_Hparams, **Hparams}#Merge default and given Hparams
        self.device_list = device_list if device_list!=[] else [torch.device('cpu')]
        self.nk = len(parameters["subkey"]) if isinstance(parameters["subkey"], list) else 1
        self.nk_per_device = np.ceil(self.nk/len(self.device_list))
        in_features = parameters["n_samples"]
        
        #Divise the branch equally between the number of available devices
        self.multy_byte_mlps = nn.ModuleDict([[str(k), MLP_byte({"n_samples":in_features, "num_classes":parameters["num_classes"]}, device_list=self.device_list, device_idx=int(k//self.nk_per_device))] for k in range(self.nk)])

    def forward(self, x):
        #Divise the branch equally between the number of available devices
        x = [self.multy_byte_mlps[str(k)](x) for k in range(self.nk)]
        return x
    
    def fit(self,*args, **kwargs):
        if "Hparams" in kwargs.keys():
            kwargs["Hparams"] = {**kwargs["Hparams"], **self.Hparams}
        else:
            kwargs["Hparams"] = self.Hparams
        return multi_fit(self, *args, **kwargs)

def multi_criterion(loss_func,outputs,labels):
  losses = 0
  m = nn.LogSoftmax(dim=1)
  for k in range(labels.shape[1]):
    losses += loss_func(m(outputs[k]), labels[:,k].long().to(device))
  return losses/labels.shape[1]

def multi_fit(model, parameters, trainset, valset, Hparams={}, callbacks=[Callbacks.Callback()], verbose=0, *args, **kwargs):
    default_Hparams = {"optimizer":"adam", "weight_init":"he_uniform", "loss_function":"crossentropy", "lr":1e-3, "bs":128, "num_epochs":100}
    Hparams = {**default_Hparams, **Hparams}
    logging.info("fit Hparams: {}".format(Hparams))
    weight_init_func = {"he_uniform":Utils.he_uniform_init_weights, "zero":Utils.zero_init_weights}[Hparams['weight_init'].lower()]
    optimizer_func = {"adam":optim.Adam, "radam":optim.RAdam, "rmsprop":optim.RMSprop, "sgd":optim.SGD}[Hparams['optimizer'].lower()]
    loss_function = {"crossentropy":nn.CrossEntropyLoss(), "mse":nn.MSELoss()}[Hparams['loss_function'].lower()] if isinstance(Hparams['loss_function'], str) else Hparams['loss_function']
    learning_rate = float(Hparams['lr'])
    batch_size = int(Hparams['bs'])
    epochs = int(Hparams['num_epochs'])

    #Create loaders
    trainloader = torch.utils.data.DataLoader(trainset, batch_size=batch_size, shuffle=True, pin_memory=True, num_workers=4)
    valloader = torch.utils.data.DataLoader(valset, batch_size=parameters['val_size'], shuffle=True, pin_memory=True, num_workers=4)#no need for gradient descent

    # model.to(device)
    model.apply(weight_init_func)
    optimizer = optimizer_func(model.parameters(), lr=learning_rate, weight_decay=1e-5)
    # lrsc = optim.lr_scheduler.CosineAnnealingLR(optimizer=optimizer,T_max=100)

    history = {'loss':torch.zeros((epochs), device=device), 'accuracy':torch.zeros((epochs), device=device), 'val_loss':torch.zeros((epochs), device=device), 'val_accuracy':torch.zeros((epochs), device=device)}
    model.train()
    skip = [cb.on_train_begin() for cb in callbacks]
    for epoch in range(epochs):
        start_t = time.time()
        skip = [cb.on_epoch_begin() for cb in callbacks]
        for i, (data, labels) in enumerate(trainloader):
            skip = [cb.on_batch_begin() for cb in callbacks]
            inputs, labels = data.to(device), labels.to(device)
            # forward + backward + optimize
            optimizer.zero_grad()
            outputs = model(inputs)
            loss = multi_criterion(loss_function, outputs, labels)#Multi-task aggregation
            loss.backward()
            skip = [cb.on_loss_end() for cb in callbacks]
            skip = [cb.on_step_begin() for cb in callbacks]
            optimizer.step()
            skip = [cb.on_step_end() for cb in callbacks]
            history['loss'][epoch] += loss.detach()
            history['accuracy'][epoch] += sum([(outputs[k].argmax(dim=1) == labels[:,k].flatten()).sum() for k in range(model.nk)])/model.nk
        history['loss'][epoch] /= (i+1)
        history['accuracy'][epoch] /= parameters['training_size']

        #Validation pass
        with torch.no_grad():
            for i, (data, labels) in enumerate(valloader):
                inputs, labels = data.to(device), labels.to(device)
                # forward only
                outputs = model(inputs)
                loss = multi_criterion(loss_function, outputs, labels)#Multi-task aggregation
                history['val_loss'][epoch] += loss
                history['val_accuracy'][epoch] += sum([(outputs[k].argmax(dim=1) == labels[:,k]).sum() for k in range(model.nk)])/model.nk
            history['val_loss'][epoch] /= (i+1)
            history['val_accuracy'][epoch] /= parameters['val_size']
            skip = [cb.on_epoch_end(model=model, epoch=epoch, history=history, last_epoch=epochs) for cb in callbacks]

        if verbose > 0:
            logger.info("Epoch {}/{} - {:.3f}s - loss: {:.4f} - accuracy: {:.4f} - val_loss: {:.4f} - val_accuracy: {:.4f}".format(epoch+1, epochs, time.time()-start_t, history['loss'][epoch], history['accuracy'][epoch], history['val_loss'][epoch], history['val_accuracy'][epoch]))
        # lrsc.step()
    if verbose == 0:
        logger.info("\nEpoch {}/{} - {:.3f}s - loss: {:.4f} - accuracy: {:.4f} - val_loss: {:.4f} - val_accuracy: {:.4f}".format(epoch+1, epochs, time.time()-start_t, history['loss'][epoch], history['accuracy'][epoch], history['val_loss'][epoch], history['val_accuracy'][epoch]))
    skip = [cb.on_train_end(model) for cb in callbacks]

    #detach and Save metrics in reults folder
    np_history = {}
    for key in history.keys():
        np_history[key] = history[key].detach().cpu().numpy()
    return np_history


# eSCAPyDL - a Deep Learning Side-Channel Analysis Python Framework
# Copyright (C) 2023  Weissbart Léo

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import torch
import numpy as np

padd = b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x10\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x02'

def to_s(state_bytes):
    s = [0] * 12
    j = 0
    for i in range(12):
        s[i] = state_bytes[j] << 0
        j += 1
        s[i] |= state_bytes[j] << 8
        j += 1
        s[i] |= state_bytes[j] << 16
        j += 1
        s[i] |= state_bytes[j] << 24
        j += 1
    return s

def to_state_bytes(s):
    state_bytes = bytearray(48)
    j = 0
    for i in range(12):
        state_bytes[j] = (s[i] >> 0) & 0xFF
        j += 1
        state_bytes[j] = (s[i] >> 8) & 0xFF
        j += 1
        state_bytes[j] = (s[i] >> 16) & 0xFF
        j += 1
        state_bytes[j] = (s[i] >> 24) & 0xFF
        j += 1
    return state_bytes

def first_round_linear_layer(state_bytes):
    s = to_s(state_bytes)

    # Only round zero is done here
    e = [0, 0, 0, 0]

    #Theta
    for i in range(4):
        r = s[i] ^ s[i + 4] ^ s[i + 8]
        e[i] = (r >> 18) | (r << 14) & 0xFFFF_FFFF
        r = e[i]
        e[i] ^= (r >> 9) | (r << 23) & 0xFFFF_FFFF
    for i in range(12):
        s[i] ^= e[(i - 1) & 3]

    #Rho west
    s[7], s[4] = s[4], s[7]
    s[7], s[5] = s[5], s[7]
    s[7], s[6] = s[6], s[7]
    for i in range(4):
        s[i+8] = (s[i+8] >> 21) | (s[i+8] << 11) & 0xFFFF_FFFF

    state = to_state_bytes(s)
    return state

def first_round_linear_layer_with_iota(state_bytes):
    s = to_s(state_bytes)

    # Only round zero is done here
    e = [0, 0, 0, 0]

    #Theta
    for i in range(4):
        r = s[i] ^ s[i + 4] ^ s[i + 8]
        e[i] = (r >> 18) | (r << 14) & 0xFFFF_FFFF
        r = e[i]
        e[i] ^= (r >> 9) | (r << 23) & 0xFFFF_FFFF
    for i in range(12):
        s[i] ^= e[(i - 1) & 3]

    #Rho west
    s[7], s[4] = s[4], s[7]
    s[7], s[5] = s[5], s[7]
    s[7], s[6] = s[6], s[7]
    for i in range(4):
        s[i+8] = (s[i+8] >> 21) | (s[i+8] << 11) & 0xFFFF_FFFF

    #Iota
    s[0] ^= 0x058

    state = to_state_bytes(s)

    return state

def np_first_round_linear_layer_with_iota(s):

    # Only round zero is done here
    e = [0, 0, 0, 0]

    #Theta
    for i in range(4):
        r = s[:,i] ^ s[:,i + 4] ^ s[:,i + 8]
        e[i] = (r >> 18) | (r << 14) & 0xFFFF_FFFF
        r = e[i]
        e[i] ^= (r >> 9) | (r << 23) & 0xFFFF_FFFF
    for i in range(12):
        s[:,i] ^= e[(i - 1) & 3]

    #Rho west
    tmp = s[:,7].copy()
    s[:,7] = s[:,4]
    s[:,4] = tmp
    tmp = s[:,7].copy()
    s[:,7] = s[:,5]
    s[:,5] = tmp
    tmp = s[:,7].copy()
    s[:,7] = s[:,6]
    s[:,6] = tmp
    for i in range(4):
        s[:,i+8] = (s[:,i+8] >> 21) | (s[:,i+8] << 11) & 0xFFFF_FFFF

    #Iota
    s[:,0] ^= 0x058

    return s

def getBit(state_bytes,x,y,z):
    #input is a bytearray of 48 bytes
    s = to_s(state_bytes)
    lane = s[x+4*y]
    return (lane >> z)&1

def getBitlane(s,x,y,z):
    #input is a numpy array of 12 lanes
    lane = s[:,x+4*y]
    return ((lane >> z)&1).astype(np.uint8)

def getBit_s(s,x,y,z):
    #input is a bytearray of 48 bytes
    # Same as previous but without conversion to state_bytes (speedup factor 13x)
    lane = s[4*(x+4*y):4*(x+4*y+1)]
    return (np.frombuffer(lane,dtype=np.uint32)[0] >> z)&1

def getColumn(state_bytes,x,z):
    return (getBit_s(state_bytes,x,0,z),getBit_s(state_bytes,x,1,z),getBit_s(state_bytes,x,2,z))
    


##
def chi3(a,b,c):
    return (a^((b^1)&c),b^((c^1)&a),c^((a^1)&b))

def rho_east_p(x,y,z):
    if(y==0):
        return (x,y,z)
    else:
       if(y==1):
           return (x,y,(z+1)%32)
       else:
           return ((x+2)%4,y,(z+8)%32)
       
def compute_intermediates(nonce, keyguess, targetbit):
    # Does it make sense to have 'nonceState' and 'nonceBeforeChi' inside trace_parameters?
    x,z = targetbit//32, targetbit%32
    # build nonce state and compute before Chi
    nonceState = bytearray(48)
    # put the nonce
    for j, v in enumerate(nonce):
        nonceState[16+j] ^= v
    # padd
    nonceState[32] ^= 0x10
    nonceState[33] ^= 0x01
    nonceState[47] ^= 0x02
    nonceBeforeChi = first_round_linear_layer_with_iota(nonceState)
    m0,m1,m2 = getColumn(nonceBeforeChi,x,z)

    ## target column is defined by the coordinates of the plane (x,z), with x index of the lane and z index of the of the bit in the lane
    fk0 = (keyguess>>3)&1 # this is the fixed key
    k0 = keyguess & 1 # this is before chi
    k1 = (keyguess>>1)&1
    k2 = (keyguess>>2)&1
    bits_aft = chi3(m0^k0,m1^k1,m2^k2)
    bits_bef = []
    bits_bef.append(fk0)
    (xa,ya,za) = rho_east_p(x,1,z)
    bits_bef.append(getBit_s(nonceState,xa,ya,za))      
    (xa,ya,za) = rho_east_p(x,2,z)              
    bits_bef.append(getBit_s(nonceState,xa,ya,za))
    return (bits_bef[0] ^ bits_aft[0]) + (bits_bef[1] ^ bits_aft[1]) + (bits_bef[2] ^ bits_aft[2])

def intermediate_value(keyguess, trace_parameters):
    # Does it make sense to have 'nonceState' and 'nonceBeforeChi' inside trace_parameters?
    x,z = trace_parameters['subkey']//32, trace_parameters['subkey']%32
    ## target column is defined by the coordinates of the plane (x,z), with x index of the lane and z index of the of the bit in the lane
    fk0 = (keyguess>>3)&1 # this is the fixed key
    k0 = keyguess & 1 # this is before chi
    k1 = (keyguess>>1)&1
    k2 = (keyguess>>2)&1

    if isinstance(trace_parameters['nonce'], np.ndarray) or isinstance(trace_parameters['nonce'], torch.Tensor):
        if len(trace_parameters['nonce'].shape) > 1:
            fk0 = fk0.numpy()
            k0 = k0.numpy()
            k1 = k1.numpy()
            k2 = k2.numpy()
            n_nonces = trace_parameters['nonce'].shape[0]
            nonceState = np.hstack((np.zeros((n_nonces,16), dtype=np.uint8), trace_parameters['nonce'], np.zeros((n_nonces,48-16-trace_parameters['nonce'].shape[1]), dtype=np.uint8)))
            nonceState ^= np.frombuffer(padd, dtype=np.uint8)
            nonceState = np.frombuffer(bytes(nonceState), dtype=np.uint32).reshape((n_nonces,12))
            nonceBeforeChi = nonceState.copy()
            nonceBeforeChi = np_first_round_linear_layer_with_iota(nonceBeforeChi)
            m0, m1, m2 = getBitlane(nonceBeforeChi,x,0,z), getBitlane(nonceBeforeChi,x,1,z), getBitlane(nonceBeforeChi,x,2,z)
            bits_aft = chi3(m0^k0,m1^k1,m2^k2)
            bits_bef = []
            bits_bef.append(fk0)
            (xa,ya,za) = rho_east_p(x,1,z)
            bits_bef.append(getBitlane(nonceState,xa,ya,za))      
            (xa,ya,za) = rho_east_p(x,2,z)              
            bits_bef.append(getBitlane(nonceState,xa,ya,za))
            # return torch.ByteTensor((bits_bef[0] ^ bits_aft[0]) + (bits_bef[1] ^ bits_aft[1]) + (bits_bef[2] ^ bits_aft[2]))
            return torch.ByteTensor(((bits_bef[0] ^ bits_aft[0])<<2) + ((bits_bef[1] ^ bits_aft[1])<<1) + (bits_bef[2] ^ bits_aft[2]))
    
    # build nonce state and compute before Chi
    nonceState = bytearray(16) + bytearray(trace_parameters['nonce'].astype(np.uint8)) + bytearray(48-16-len(trace_parameters['nonce']))
    nonceState = (np.frombuffer(bytes(nonceState), dtype=np.uint32) ^ np.frombuffer(padd, dtype=np.uint32)).tobytes()
    nonceBeforeChi = first_round_linear_layer_with_iota(nonceState)
    m0,m1,m2 = getColumn(nonceBeforeChi,x,z)
    bits_aft = chi3(m0^k0,m1^k1,m2^k2)
    bits_bef = []
    bits_bef.append(fk0)
    (xa,ya,za) = rho_east_p(x,1,z)
    bits_bef.append(getBit_s(nonceState,xa,ya,za))      
    (xa,ya,za) = rho_east_p(x,2,z)              
    bits_bef.append(getBit_s(nonceState,xa,ya,za))
    return ((bits_bef[0] ^ bits_aft[0])<<2) + ((bits_bef[1] ^ bits_aft[1])<<1) + (bits_bef[2] ^ bits_aft[2])

def select_subkey(i, key):
    # return the subkey 
    x,z = i//32, i%32
    # build key state and compute before Chi
    keyInit = bytearray(48)
    # put the key
    for j, v in enumerate(key):
        keyInit[j] ^= v
    keyBeforeChi = first_round_linear_layer(keyInit)
    ck0,ck1,ck2 = getColumn(keyBeforeChi,x,z)
    return (getBit_s(keyInit,x,0,z)<<3) ^ (ck2<<2) ^ (ck1<<1) ^ ck0


################################################
## Chi output without HD
################################################


def intermediate_value_chi(keyguess, trace_parameters):
    # Does it make sense to have 'nonceState' and 'nonceBeforeChi' inside trace_parameters?
    x,z = trace_parameters['subkey']//32, trace_parameters['subkey']%32
    ## target column is defined by the coordinates of the plane (x,z), with x index of the lane and z index of the of the bit in the lane
    # fk0 = (keyguess>>3)&1 # this is the fixed key
    k0 = keyguess & 1 # this is before chi
    k1 = (keyguess>>1)&1
    k2 = (keyguess>>2)&1

    if isinstance(trace_parameters['nonce'], np.ndarray) or isinstance(trace_parameters['nonce'], torch.Tensor):
        if len(trace_parameters['nonce'].shape) > 1:
            # fk0 = fk0.numpy()
            k0 = k0.numpy()
            k1 = k1.numpy()
            k2 = k2.numpy()
            n_nonces = trace_parameters['nonce'].shape[0]
            assert trace_parameters['nonce'].shape[1] == 16
            nonceState = np.hstack((np.zeros((n_nonces,16), dtype=np.uint8), trace_parameters['nonce'], np.zeros((n_nonces,48-16-trace_parameters['nonce'].shape[1]), dtype=np.uint8)))
            nonceState ^= np.frombuffer(padd, dtype=np.uint8)
            nonceState = np.frombuffer(bytes(nonceState), dtype=np.uint32).reshape((n_nonces,12))
            nonceBeforeChi = nonceState.copy()
            nonceBeforeChi = np_first_round_linear_layer_with_iota(nonceBeforeChi)
            m0, m1, m2 = getBitlane(nonceBeforeChi,x,0,z), getBitlane(nonceBeforeChi,x,1,z), getBitlane(nonceBeforeChi,x,2,z)
            bits_aft = chi3(m0^k0,m1^k1,m2^k2)
            # bits_bef = []
            # bits_bef.append(fk0)
            # (xa,ya,za) = rho_east_p(x,1,z)
            # bits_bef.append(getBitlane(nonceState,xa,ya,za))      
            # (xa,ya,za) = rho_east_p(x,2,z)              
            # bits_bef.append(getBitlane(nonceState,xa,ya,za))
            # return torch.ByteTensor((bits_bef[0] ^ bits_aft[0]) + (bits_bef[1] ^ bits_aft[1]) + (bits_bef[2] ^ bits_aft[2]))
            return torch.ByteTensor((bits_aft[0]) + (bits_aft[1]) + (bits_aft[2]))
    
    # build nonce state and compute before Chi
    assert len(bytearray(trace_parameters['nonce'].astype(np.uint8))) == 16
    nonceState = bytearray(16) + bytearray(trace_parameters['nonce'].astype(np.uint8)) + bytearray(48-16-len(trace_parameters['nonce']))
    nonceState = (np.frombuffer(bytes(nonceState), dtype=np.uint32) ^ np.frombuffer(padd, dtype=np.uint32)).tobytes()
    nonceBeforeChi = first_round_linear_layer_with_iota(nonceState)
    m0,m1,m2 = getColumn(nonceBeforeChi,x,z)
    bits_aft = chi3(m0^k0,m1^k1,m2^k2)
    # bits_bef = []
    # bits_bef.append(fk0)
    # (xa,ya,za) = rho_east_p(x,1,z)
    # bits_bef.append(getBit_s(nonceState,xa,ya,za))      
    # (xa,ya,za) = rho_east_p(x,2,z)              
    # bits_bef.append(getBit_s(nonceState,xa,ya,za))
    # return (bits_bef[0] ^ bits_aft[0]) + (bits_bef[1] ^ bits_aft[1]) + (bits_bef[2] ^ bits_aft[2])
    return (bits_aft[0]) + (bits_aft[1]) + (bits_aft[2])

def select_subkey_chi(i, key):
    # return the subkey 
    x,z = i//32, i%32
    # build key state and compute before Chi
    keyInit = bytearray(48)
    # put the key
    for j, v in enumerate(key):
        keyInit[j] ^= v
    keyBeforeChi = first_round_linear_layer(keyInit)
    ck0,ck1,ck2 = getColumn(keyBeforeChi,x,z)
    return (ck2<<2) ^ (ck1<<1) ^ ck0#(getBit_s(keyInit,x,0,z)<<3) ^ 


# eSCAPyDL - a Deep Learning Side-Channel Analysis Python Framework
# Copyright (C) 2023  Weissbart Léo

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os, sys
import shutil
import json
import logging 
logging.basicConfig(format='%(message)s')
logger = logging.getLogger()
logger.setLevel(logging.INFO)
import time, datetime
import argparse
import inspect

import torch
from torch.utils import tensorboard
import Dataset
import Callbacks
import Models
import Utils

def main():
    #Parse arguments
    #TODO: remove default dataset and model and default to help
    parser = argparse.ArgumentParser(description='eSCAPyDL: Deep Leaning SCA PyTorch framework. Training models for tracesets with SCA metrics.')
    parser.add_argument('-d', '--dataset', help='The dataset to use for training. (All datasets have different functions for intermediate value and power model).', default="dpav4")
    parser.add_argument('-m', '--model', help='Type of NN to use (choices: MLP, CNN) CNN is the onetrace implementation.', default="MLP")
    parser.add_argument('-j', '--json', help='Input data from json file')
    parser.add_argument('-a', '--timestamp', help='Timestamp of the experiment', default=time.strftime("%Y%m%d_%H%M%S",time.localtime()))
    parser.add_argument('-v', '--verbose', help='Verbose level', action='count', default=0)
    parser.add_argument('-l', '--log', help='Log into a file at path', default='')
    args = parser.parse_args()
    available_models = [x[0] for x in inspect.getmembers(Models, inspect.isclass)]
    assert args.model in available_models, "Model {} not found, available models are: {}".format(args.model, available_models)
    if args.verbose == 1:
        logger.setLevel(logging.DEBUG)
        logging.getLogger("Dataset").setLevel(logging.DEBUG)
        logging.getLogger("Callbacks").setLevel(logging.DEBUG)
        logging.getLogger("models").setLevel(logging.DEBUG)
        logging.getLogger("utils").setLevel(logging.DEBUG)
    if args.log != '':
        file_handler = logging.FileHandler(args.log, mode='w', encoding='utf-8')
        file_handler.setFormatter(logging.Formatter('%(asctime)s %(levelname)s %(message)s', datefmt='%d/%m/%Y %H:%M:%S'))
        logger.addHandler(file_handler)
        logger.setLevel(logging.INFO)
    __trs_dataset__ = args.dataset
    __DL_type__ = args.model
    if args.json == "None":
        args.json = None
    if args.json != None:
        try:
            with open(args.json, 'r') as f:
                json_data = json.load(f)
        except FileNotFoundError:
            logger.error("Json File '{}' not found".format(args.json))
            exit(1)
    else:
        json_data = {}

    #Main program
    try:
        start_time = datetime.datetime.now().replace(microsecond=0)

        # Create folders
        filebase = "{}/{}/{}".format(__trs_dataset__, __DL_type__, args.timestamp)
        root                = os.getcwd() + '/experiments/' + filebase + '/'
        os.makedirs(root)
        model_folder        = root + 'models/'
        last_model_folder   = root + 'last_models/'
        history_folder      = root + 'history/'
        result_folder       = root + 'results/'
        for folder in [model_folder,model_folder,last_model_folder,history_folder,result_folder]:
            try:
                os.mkdir(folder)
            except:
                pass

        #load traceset
        dataset = Dataset.Dataset(__trs_dataset__, __DL_type__, json_data=json_data)
        parameters = dataset.get_parameters()
        Hparams = dataset.get_model_parameters()
        callbacks_params = dataset.get_callbacks()
        #Load traces
        trace_data = dataset.traceset()
        trainset = torch.utils.data.TensorDataset(torch.Tensor(trace_data["x_train"]), torch.Tensor(trace_data["m_train"]["label"]))
        valset = torch.utils.data.TensorDataset(torch.Tensor(trace_data["x_val"]), torch.Tensor(trace_data["m_val"]["label"]))

        # Create and train model
        try:
            model = eval("Models."+__DL_type__)(parameters, Hparams=Hparams)
        except AttributeError:
            raise Exception("module 'Models' has no attribute '{}', available attributes are: {}".format(__DL_type__, [x[0] for x in inspect.getmembers(Models, inspect.isclass)]))
        
        #Print train session summary for debug
        logger.info("Dataset: {}".format(__trs_dataset__))
        logger.info("Model: {}".format(__DL_type__))
        logger.info("Parameters: {}".format(dataset.parameters))
        logger.info("Model parameters: {}".format(Hparams))
        logger.info(model)
        # logging.debug(summary(model, trace_data["x_train"][0].shape, device='cpu'))

        #Create callbacks
        callbacks = Callbacks.init_callbacks(callbacks_params, 
                                   traceset=trace_data, 
                                   Hparams=Hparams, 
                                   parameters=parameters, 
                                   exp_path=result_folder, 
                                   writer_path='runs/'+filebase)

        #Train model
        history = model.fit(trainset=trainset, 
                            valset=valset, 
                            parameters=parameters, 
                            Hparams = Hparams,
                            callbacks=callbacks,
                            verbose=1,
                        )
        #Save results
        Utils.save_model_weights(model, os.path.join(last_model_folder, 'last_model.pt'))
        Utils.save_history(history, history_folder + "history")
        Utils.save_db(root+'{}.sqlite'.format("experiment_db"), __trs_dataset__, dataset.parameters, model, history)

        stop_time = datetime.datetime.now().replace(microsecond=0)
        logger.info("----------------------\n  The operation was \ncompleted successfully\n Clear time {}\n----------------------".format(stop_time-start_time))
    except:
        # Create handler to remove uncessful experiment folder tree if program crashed
        logger.warning("------------------\nException caught. Cleaning experiment folder..\n------------------")
        if os.path.isdir(root):
            shutil.rmtree(root)
        if os.path.isdir('runs/'+filebase):
            shutil.rmtree('runs/'+filebase)
        logger.exception('')
    finally:
        logger.info("--------------------\nEnd of main reached!")
    logger.info("--------------------")

if __name__ == "__main__":
    sys.exit(main())


import os
import torch
import Utils
import helpers.ascon_helper as ascon

sbox_chunk_size = 8
parameters = {
    "dataset_name"        : "/mnt/D_drive/leo/eSCAPyDL/data/ASCON_P3_CPA_all_mlpsca_ready.trs",
    "dataset_size"        : 600000,
    "training_size"       : 500000,
    "val_size"            : 100000,
    "subkey"              : 0,
    "num_key_hypotheses"  : 2**(sbox_chunk_size),
    "n_samples"           : 2708,
    "r"                   : 0.8,
    "n_repeat"            : 50,
    "shuffle"             : True,
    "ge_rate"             : 5,
    "chunk_size"          : sbox_chunk_size,
}

model = {
    "num_epochs"          : 500,
    "loss_function"       : Utils.FocalLoss(),
    "lr"                  : 1e-03,
    "bs"                  : 512,
}

callbacks = {
    "modelcheckpoint":{
        "period":-1,
    },
    "guessingentropy":{
        "ge_rate":5,
        "n_repeat":5,
        "r":0.6,
    },
    "tensorboard":{
        "training_accuracy":True,
        "validation_accuracy":True,
        "training_loss":True,
        "validation_loss":True,
    },
}

def get_attack_parameters(traceset, subkey, index):
    """Returns a dictionary of parameters to be used to search for attack .
    To be used before calling intermediate_value().

    :param traceset: Traceset to use
    :type traceset: dict
    :param subkey: Subkey to attack
    :type subkey: int
    :param index: Index of the traceset to use
    :type index: int or slice
    :return: Dictionary of parameters
    :rtype: dict
    """        
    return {'nonce0':traceset["nonce0"][index],
            'nonce1':traceset["nonce1"][index],
            'subkey':subkey}

def get_known_key(traceset, subkey, index):
    return ascon.select_subkey_sbox_chunk(subkey, traceset['key'][index], chunk_size=parameters["chunk_size"])

def intermediate_value(keyguess, trace_parameters):
    """Return the intermediate value of a key in the SBox . 
    nonce0 and nonce1 should be byte arrays of size (8) as they both are half of the nonce length (16 bytes).
    k should be a 3-bit value as an integer. This bit corresponds to the part of the key concerned in the attack given subkey.

    :param trace_parameters: Dictionary containing the trace parameters for intermediate value computation
    :type trace_parameters: dict
    :return: Intermediate value
    :rtype: Any
    """        
    return ascon.intermediate_value_sbox_chunk(keyguess,trace_parameters, chunk_size=parameters["chunk_size"])



import torch
import optuna
import numpy as np
import sys,os
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../escapydl/')))

## Test functions in Optuna_models.py
def test_is_function_local():
    from escapydl.Optuna_models import is_function_local
    assert is_function_local(lambda x: x) == False, "expected output: {}, actual output: {}".format(False, is_function_local(lambda x: x))

def test_make_MLP():
    from escapydl.Optuna_models import objective_hp
    __DL_type__ = "make_MLP"
    study_name = "unit_test_make_MLP"
    db_name = "sqlite:///unit_test_make_MLP.db"

    parameters = {
        "n_samples": 10,
        "num_classes": 2,
        "training_size": 10,
        "val_size": 2,
    }

    trace_data = {"x_train":torch.randn(10,10),
                  "x_val":torch.randn(2,10),
                  "m_train":{"label":torch.randint(0, 2, (10,))},
                  "m_val":{"label":torch.randint(0, 2, (2,))},
                  }

    hp_search_s = {
        "activation_function_s": ['relu', 'tanh', 'selu'],
        "epochs": [1,1,1], 
        "bs": [32,128,32],
        "loss_function_s": ["crossentropy"],
        "n_neurons_s": [10,20,10], 
        "n_layers_s": [1,1,1], 
        "kernel_size_s": [3,8,1],
        "n_conv_block_s": [1,1,1],
        "optimizer_s": ["Adam"], 
        "lr_s": [1e-3,1e-3,'log']
    }

    metric, direction = "None", "minimize"
    objective = objective_hp(parameters=parameters,
                        hp_search_s=hp_search_s,
                        trace_data=trace_data,
                        verbose=1,
                        model_name=__DL_type__,
                        metric=metric
                    )
    pruner = optuna.pruners.PatientPruner(optuna.pruners.MedianPruner(n_startup_trials=5, n_warmup_steps=5, interval_steps=1), patience=30)
    
    study = optuna.create_study(sampler=optuna.samplers.TPESampler(), 
                                direction=direction, 
                                pruner=pruner, 
                                study_name=study_name, 
                                storage=db_name)
    study.optimize(objective, n_trials=1, timeout=None)
    assert os.path.exists("unit_test_make_MLP.db") == True, "expected output: {}, actual output: {}".format(True, os.path.exists("unit_test_make_MLP.db"))
    os.remove("unit_test_make_MLP.db")

def test_make_CNN():
    from escapydl.Optuna_models import objective_hp
    __DL_type__ = "make_CNN"
    study_name = "unit_test_make_CNN"
    db_name = "sqlite:///unit_test_make_CNN.db"

    parameters = {
        "n_samples": 10,
        "num_classes": 2,
        "training_size": 10,
        "val_size": 2,
    }

    trace_data = {"x_train":torch.randn(10,1,10),
                  "x_val":torch.randn(2,1,10),
                  "m_train":{"label":torch.randint(0, 2, (10,))},
                  "m_val":{"label":torch.randint(0, 2, (2,))},
                  }

    hp_search_s = {
        "activation_function_s": ['relu', 'tanh', 'selu'],
        "epochs": [1,1,1], 
        "bs": [32,128,32],
        "loss_function_s": ["crossentropy"],
        "n_neurons_s": [10,20,10], 
        "n_layers_s": [1,1,1], 
        "kernel_size_s": [3,8,1],
        "n_conv_block_s": [1,1,1],
        "optimizer_s": ["Adam"], 
        "lr_s": [1e-3,1e-3,'log']
    }

    metric, direction = "None", "minimize"
    objective = objective_hp(parameters=parameters,
                        hp_search_s=hp_search_s,
                        trace_data=trace_data,
                        verbose=1,
                        model_name=__DL_type__,
                        metric=metric
                    )
    pruner = optuna.pruners.PatientPruner(optuna.pruners.MedianPruner(n_startup_trials=5, n_warmup_steps=5, interval_steps=1), patience=30)
    
    study = optuna.create_study(sampler=optuna.samplers.TPESampler(), 
                                direction=direction, 
                                pruner=pruner, 
                                study_name=study_name, 
                                storage=db_name)
    study.optimize(objective, n_trials=1, timeout=None)
    assert os.path.exists("unit_test_make_CNN.db") == True, "expected output: {}, actual output: {}".format(True, os.path.exists("unit_test_make_CNN"))
    os.remove("unit_test_make_CNN.db")

def test_make_CNN_simpler():
    from escapydl.Optuna_models import objective_hp
    __DL_type__ = "make_CNN_simpler"
    study_name = "unit_test_make_CNN_simpler"
    db_name = "sqlite:///unit_test_make_CNN_simpler.db"

    parameters = {
        "n_samples": 1000,
        "num_classes": 2,
        "training_size": 10,
        "val_size": 2,
    }

    trace_data = {"x_train":torch.randn(10,1,1000),
                  "x_val":torch.randn(2,1,1000),
                  "m_train":{"label":torch.randint(0, 2, (10,))},
                  "m_val":{"label":torch.randint(0, 2, (2,))},
                  }

    hp_search_s = {
        "activation_function_s": ['relu', 'tanh', 'selu'],
        "epochs": [1,1,1], 
        "bs": [32,128,32],
        "loss_function_s": ["crossentropy"],
        "n_neurons_s": [10,20,10], 
        "n_layers_s": [1,1,1], 
        "kernel_size_s": [3,8,1],
        "n_conv_block_s": [1,1,1],
        "optimizer_s": ["Adam"], 
        "lr_s": [1e-3,1e-3,'log']
    }

    metric, direction = "None", "minimize"
    objective = objective_hp(parameters=parameters,
                        hp_search_s=hp_search_s,
                        trace_data=trace_data,
                        verbose=1,
                        model_name=__DL_type__,
                        metric=metric
                    )
    pruner = optuna.pruners.PatientPruner(optuna.pruners.MedianPruner(n_startup_trials=5, n_warmup_steps=5, interval_steps=1), patience=30)
    
    study = optuna.create_study(sampler=optuna.samplers.TPESampler(), 
                                direction=direction, 
                                pruner=pruner, 
                                study_name=study_name, 
                                storage=db_name)
    study.optimize(objective, n_trials=1, timeout=None)
    assert os.path.exists("unit_test_make_CNN_simpler.db") == True, "expected output: {}, actual output: {}".format(True, os.path.exists("unit_test_make_CNN_simpler"))
    os.remove("unit_test_make_CNN_simpler.db")

def test_objective_hp():
    assert True, "already covered"


if __name__ == "__main__":
    test_is_function_local()
    print("test_is_function_local passed")

    test_make_MLP()
    print("test_make_MLP passed")

    test_make_CNN()
    print("test_make_CNN passed")

    test_make_CNN_simpler()
    print("test_make_CNN_simpler passed")

    test_objective_hp()
    print("test_objective_hp passed")


import torch
import numpy as np
import sys,os
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

## Test functions in Utils.py
def test_he_uniform_init_weights():
    from escapydl.Utils import he_uniform_init_weights    
    m = torch.nn.Linear(10, 10)
    he_uniform_init_weights(m)
    assert m.weight.shape == (10, 10), "expected output: {}, actual output: {}".format((10, 10), m.weight.shape)
    assert m.bias.shape == (10,), "expected output: {}, actual output: {}".format((10,), m.bias.shape)

def test_zero_init_weights():
    from escapydl.Utils import zero_init_weights
    m = torch.nn.Linear(10, 10)
    zero_init_weights(m)
    assert (m.weight == torch.zeros((10, 10))).all(), "expected output: {}, actual output: {}".format(torch.zeros((10, 10)), m.weight)
    assert (m.bias == torch.zeros((10,))).all(), "expected output: {}, actual output: {}".format(torch.zeros((10,)), m.bias)

def test_FocalLoss():
    from escapydl.Utils import FocalLoss
    loss = FocalLoss()
    output = loss(torch.Tensor([[0.1, 0.1, 0.1],[0.1, 0.1, 0.1],[0.1, 0.1, 0.1]]), torch.ByteTensor([0, 0, 1]))
    assert output.dim() == 0, "expected output: {}, actual output: {}".format(0, output.dim())

def test_RankLoss():
    from escapydl.Utils import RankLoss
    loss = RankLoss()
    output = loss(torch.Tensor([[0.1, 0.1, 0.1],[0.1, 0.1, 0.1],[0.1, 0.1, 0.1]]), torch.ByteTensor([0, 0, 1]))
    assert output.shape == (3,2), "expected output: {}, actual output: {}".format((3,2), output.shape)

def test_LocallyConnected2d():
    from escapydl.Utils import LocallyConnected2d
    m = LocallyConnected2d(1, 1, 1, 1, 1, bias=True)
    assert m.weight.shape == (1, 1, 1, 1, 1, 1), "expected output: {}, actual output: {}".format((1, 1, 1, 1, 1, 1), m.weight.shape)
    assert m.bias.shape == (1, 1, 1, 1), "expected output: {}, actual output: {}".format((1, 1, 1, 1), m.bias.shape)
    input = torch.Tensor([[[[1]]]])
    output = m(input)
    assert output.shape == (1, 1, 1, 1), "expected output: {}, actual output: {}".format((1, 1, 1, 1), output.shape)

def test_LocallyConnected1d():
    from escapydl.Utils import LocallyConnected1d
    m = LocallyConnected1d(1, 1, 1, 1, 1, bias=True)
    assert m.weight.shape == (1, 1, 1, 1, 1), "expected output: {}, actual output: {}".format((1, 1, 1, 1, 1), m.weight.shape)
    assert m.bias.shape == (1, 1, 1), "expected output: {}, actual output: {}".format((1, 1, 1), m.bias.shape)
    input = torch.Tensor([[[1]]])
    output = m(input)
    assert output.shape == (1, 1, 1), "expected output: {}, actual output: {}".format((1, 1, 1), output.shape)

def test_NumpyArrayEncoder():
    from escapydl.Utils import NumpyArrayEncoder
    import json
    data = {
        "test": np.array([1,2,3])
    }
    output = json.dumps(data, cls=NumpyArrayEncoder)
    assert output == '{"test": [1, 2, 3]}', "expected output: {}, actual output: {}".format('{"test": [1, 2, 3]}', output)

def test_save_history():
    from escapydl.Utils import save_history
    save_history({"test": [1,2,3]}, "test_save_history")
    assert os.path.isfile("test_save_history.json"), "expected output: {}, actual output: {}".format(True, os.path.isfile("test_save_history.json"))
    os.remove("test_save_history.json")

def test_load_history():
    from escapydl.Utils import save_history, load_history
    save_history({"test": [1,2,3]}, "test_save_history")
    output = load_history("test_save_history.json")
    assert output == {"test": [1,2,3]}, "expected output: {}, actual output: {}".format({"test": [1,2,3]}, output)
    os.remove("test_save_history.json")

def test_save_model_weights():
    from escapydl.Utils import save_model_weights
    model = torch.nn.Linear(10, 10)
    save_model_weights(model, "test_save_model_weights.pth")
    assert os.path.isfile("test_save_model_weights.pth"), "expected output: {}, actual output: {}".format(True, os.path.isfile("test_save_model_weights.pth"))
    os.remove("test_save_model_weights.pth")

def test_load_model_weights():
    from escapydl.Utils import save_model_weights, load_model_weights
    model = torch.nn.Linear(10, 10)
    weights = model.state_dict()
    save_model_weights(model, "test_save_model_weights.pth")
    load_model_weights(model, "test_save_model_weights.pth")
    l_weights = model.state_dict()
    assert (weights['weight'] == l_weights['weight']).all(), "expected output: {}, actual output: {}".format(weights['weight'], l_weights['weight'])
    assert (weights['bias'] == l_weights['bias']).all(), "expected output: {}, actual output: {}".format(weights['bias'], l_weights['bias'])
    os.remove("test_save_model_weights.pth")

def test_save_model_torchscript():
    from escapydl.Utils import save_model_torchscript
    model = torch.nn.Linear(10, 10)
    save_model_torchscript(model, "test_save_model_torchscript.pth")
    assert os.path.isfile("test_save_model_torchscript.pth"), "expected output: {}, actual output: {}".format(True, os.path.isfile("test_save_model_torchscript.pth"))
    os.remove("test_save_model_torchscript.pth")

def test_load_model_torchscript():
    from escapydl.Utils import save_model_torchscript, load_model_torchscript
    model = torch.nn.Linear(10, 10)
    save_model_torchscript(model, "test_save_model_torchscript.pth")
    output = load_model_torchscript("test_save_model_torchscript.pth")
    assert isinstance(output, torch.jit.ScriptModule), "expected output: {}, actual output: {}".format(torch.jit.ScriptModule, type(output))
    os.remove("test_save_model_torchscript.pth")

def test_shuffle_numpy():
    from escapydl.Utils import shuffle_numpy
    x = np.arange(20).reshape((2,10))
    shuffle_numpy(x)
    assert x.shape == (2,10), "expected output: {}, actual output: {}".format((2,10), x.shape)

def test_shuffle_dict():
    from escapydl.Utils import shuffle_dict
    x = {
        "a": np.arange(20).reshape((2,10)),
        "b": np.arange(20).reshape((2,10))
    }
    shuffle_dict(x, shuffled_keys=["a","b"])
    assert x["a"].shape == (2,10), "expected output: {}, actual output: {}".format((2,10), x["a"].shape)
    assert x["b"].shape == (2,10), "expected output: {}, actual output: {}".format((2,10), x["b"].shape)

def test_shuffle_torch():
    from escapydl.Utils import shuffle_torch
    x = torch.arange(20).reshape((2,10))
    shuffle_torch(x)
    assert x.shape == (2,10), "expected output: {}, actual output: {}".format((2,10), x.shape)

def test_save_db():
    from escapydl.Utils import save_db
    db_filepath = "test_save_db.db"
    dataset_name = "dpav4"
    dataset_parameters = {
        "n_traces"  : 3,
        "n_samples" : 0,
        "description": "unit_test"
    }
    model = torch.nn.Linear(10, 10)
    model.name = "unit_test_linear"
    model.Hparams = {
        "loss_function": None
        }
    model.fit = lambda x,y: None
    history = {"test": [1,2,3]}
    try:
        save_db(db_filepath, dataset_name, dataset_parameters, model, history)
        assert os.path.isfile("test_save_db.db"), "expected output: {}, actual output: {}".format(True, os.path.isfile("test_save_db.db"))
        os.remove("test_save_db.db")
    except Exception as e:
        os.remove("test_save_db.db")
        raise e
    
def test_load_db():
    from escapydl.Utils import save_db, load_db
    db_filepath = "test_save_db.db"
    dataset_name = "dpav4"
    dataset_parameters = {
        "n_traces"  : 3,
        "n_samples" : 0,
        "description": "unit_test"
    }
    model = torch.nn.Linear(10, 10)
    model.name = "unit_test_linear"
    model.Hparams = {
        "loss_function": None
        }
    model.fit = lambda x,y: None
    history = {"test": [1,2,3]}
    try:
        save_db(db_filepath, dataset_name, dataset_parameters, model, history)
        output = load_db(db_filepath)
        assert output == dataset_name, "expected output: {}, actual output: {}".format(dataset_name, output)
        os.remove("test_save_db.db")
    except Exception as e:
        os.remove("test_save_db.db")
        raise e

def test_pbar():
    from escapydl.Utils import pbar
    for i in pbar(range(10)):
        pass
    assert True, "expected output: {}, actual output: {}".format(True, True)


if __name__ == "__main__":
    test_he_uniform_init_weights()
    print("test_he_uniform_init_weights passed")

    test_zero_init_weights()
    print("test_zero_init_weights passed")

    test_FocalLoss()
    print("test_FocalLoss passed")

    test_RankLoss()
    print("test_RankLoss passed")

    test_LocallyConnected2d()
    print("test_LocallyConnected2d passed")

    test_LocallyConnected1d()
    print("test_LocallyConnected1d passed")

    test_NumpyArrayEncoder()
    print("test_NumpyArrayEncoder passed")

    test_save_history()
    print("test_save_history passed")

    test_load_history()
    print("test_load_history passed")

    test_save_model_weights()
    print("test_save_model_weights passed")

    test_load_model_weights()
    print("test_load_model_weights passed")

    test_save_model_torchscript()
    print("test_save_model_torchscript passed")

    test_load_model_torchscript()
    print("test_load_model_torchscript passed")

    test_shuffle_numpy()
    print("test_shuffle_numpy passed")

    test_shuffle_dict()
    print("test_shuffle_dict passed")

    test_shuffle_torch()
    print("test_shuffle_torch passed")

    test_save_db()
    print("test_save_db passed")

    test_load_db()
    print("test_load_db passed")

    test_pbar()
    print("test_pbar passed")





import os
import torch
import Utils
import numpy as np
import helpers.ascon_helper as ascon

parameters = {
    "dataset_name"        : "data/ASCON_P2_CPA_all_mlpsca_ready.trs",
    "dataset_size"        : 510000,
    "training_size"       : 50000,
    "val_size"            : 5000,
    "subkey"              : [x for x in range(64)],
    "num_key_hypotheses"  : 4,
    "n_samples"           : 1408,
    "r"                   : 0.6,
    "n_repeat"            : 50,
    "ge_rate"             : 5,
    "shuffle"             : True,
}

model = {
    "num_epochs"          : 200,
    "loss_function"       : Utils.FocalLoss(),
    "lr"                  : 1e-03,
    "bs"                  : 512,
    "activation_function": "relu", 
    "pooling": "average",
    "optimizer": "radam",
    "weight_init": "he_uniform",
    "n_conv_block": 2,
    "kernel_size_0": 30, 
    "c_out_0": 8,
    "kernel_size_1": 5,
    "c_out_1": 16, 
#    "kernel_size_2": 6,
#    "c_out_2": 32, 
    "lr": 1e-3,  
    "n_layers": 4, 
    "n_units_l0": 500, 
    "n_units_l1": 470,
    "n_units_l2": 10,
    "n_units_l3": 480,
#    "n_units_l4": 100,
#    "n_units_l5": 100,
#    "n_units_l6": 100,
#    "n_units_l7": 100,
                           
}

callbacks = {
    "modelcheckpoint":{
        "period":-1,
    },
    "guessingentropy":{
        "ge_rate":5,
        "n_repeat":50,
        "r":0.6,
    },
    "tensorboard":{
        "training_accuracy":True,
        "validation_accuracy":True,
        "training_loss":True,
        "validation_loss":True,
    },
}

def get_attack_parameters(traceset, subkey, index):
    """Returns a dictionary of parameters to be used to search for attack .
    To be used before calling intermediate_value().

    :param traceset: Traceset to use
    :type traceset: dict
    :param subkey: Subkey to attack
    :type subkey: int
    :param index: Index of the traceset to use
    :type index: int or slice
    :return: Dictionary of parameters
    :rtype: dict
    """        
    return {'nonce0':traceset["nonce0"][index],
            'nonce1':traceset["nonce1"][index],
            'subkey':subkey}

def get_known_key(traceset, subkey, index):
    return ascon.select_subkey_sbox(subkey, traceset['key'][index])

def intermediate_value(keyguess, trace_parameters):
    """Return the intermediate value of a key in the SBox . 
    nonce0 and nonce1 should be byte arrays of size (8) as they both are half of the nonce length (16 bytes).
    k should be a 3-bit value as an integer. This bit corresponds to the part of the key concerned in the attack given subkey.

    :param trace_parameters: Dictionary containing the trace parameters for intermediate value computation
    :type trace_parameters: dict
    :return: Intermediate value
    :rtype: Any
    """        
    return ascon.intermediate_value_sbox(keyguess,trace_parameters)


# eSCAPyDL - a Deep Learning Side-Channel Analysis Python Framework
# Copyright (C) 2023  Weissbart Léo

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import numpy as np
import torch
import matplotlib.pyplot as plt
plt.set_loglevel("critical")
import logging 
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
from torch.utils import tensorboard

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
logger.debug(device)


def get_attack_parameters(traceset, subkey, index):
    """Returns a dictionary of parameters to be used to search for attack .
    To be used before calling intermediate_value().

    :param traceset: Traceset to use
    :type traceset: dict
    :param subkey: Subkey to target
    :type subkey: int
    :param index: Index of the traceset to use
    :type index: int or slice
    :return: Dictionary of parameters
    :rtype: dict
    """        
    return None

def get_known_key(traceset, subkey, index):
    """Returns the part of the key in traceset that corresponds to the Dataset and subkey
    
    :param traceset: Traceset to use
    :type traceset: dict
    :param subkey: Subkey to target
    :type subkey: int
    :param index: Index of the traceset to use
    :type index: int or slice
    :return: Known key
    :rtype: Any
    """
    return None

def intermediate_value(keyguess, trace_parameters):
    """Return the intermediate leakage value. 
    
    :param trace_parameters: Dictionary containing the trace parameters for intermediate value computation
    :type trace_parameters: dict
    :return: Intermediate value
    :rtype: Any
    """        
    return None


class Callback():
    """Base class for callbacks.
    """
    def __init__(self,*args,**kwargs): 
        self.name = None
    def on_train_begin(self,*args,**kwargs): pass
    def on_train_end(self,*args,**kwargs): pass
    def on_epoch_begin(self,*args,**kwargs): pass
    def on_epoch_end(self,*args,**kwargs): pass
    def on_batch_begin(self,*args,**kwargs): pass
    def on_batch_end(self,*args,**kwargs): pass
    def on_loss_begin(self,*args,**kwargs): pass
    def on_loss_end(self,*args,**kwargs): pass
    def on_step_begin(self,*args,**kwargs): pass
    def on_step_end(self,*args,**kwargs): pass
    def reset(self,*args,**kwargs): pass

class ModelCheckpoint(Callback):
    """Creates a callback that will be called when the epoch end of the training, and will save the model.
    """
    def __init__(self, path, period=50, *args, **kwargs):
        """ Initializes the callback

        :param path: Path to save the model
        :type path: str
        :param period: Interval of epochs between each save (default 50), set to -1 to desactivate saving after each epoch.
        :type period: int, optional
        """
        super(ModelCheckpoint, self).__init__(*args, **kwargs)
        self.name = 'modelcheckpoint'
        self.period = period
        self.path = path
        self.exp_path = path
        os.makedirs(self.exp_path, exist_ok=True)
    
    def reset(self, *args, **kwargs):
        self.period = kwargs['period'] if 'period' in kwargs.keys() else self.period
        self.exp_path = os.path.join(self.path,str(kwargs['trial'].number)) if 'trial' in kwargs.keys() else self.path
        os.makedirs(self.exp_path, exist_ok=True)

    def on_epoch_end(self, model, epoch, *args, **kwargs):
        #Save checkpoints every 'period' epochs in a subfolder
        if self.period != -1:
            if epoch % self.period == 0 and epoch != 0:
                if 'trial' in kwargs.keys():
                    filename = os.path.join(self.exp_path, 'model_trial{}_{:03d}epochs.pt'.format(kwargs['trial'].number, epoch))
                else:
                    filename = os.path.join(self.exp_path, 'model_{:03d}epochs.pt'.format(epoch))
                torch.save(model.state_dict(), filename)
                logger.info("Epoch {}: saving model to {}".format(epoch+1, filename))

    def on_train_end(self, model, *args, **kwargs):
        #Save final model in path
        if 'trial' in kwargs.keys():
            filename = os.path.join(self.exp_path, 'model_trial{}.pt'.format(kwargs['trial'].number))
        else:
            filename = os.path.join(self.exp_path, 'model.pt')
        torch.save(model.state_dict(), filename)
        logger.info("End of training: saving model to {}".format(filename))

class TensorboardCallback(Callback):
    """Creates a callback that will be called when the epoch end of the training, and will write the loss and accuracy in tensorboard.
    """
    def __init__(self, *args, **kwargs):
        """ Initializes the callback

        :param writer: Tensorboard writer
        :type writer: tf.summary.FileWriter, optional
        """
        super().__init__(*args, **kwargs)
        #Get the writer in kwargs
        self.name = 'tensorboard'
        if 'path' in kwargs.keys():
            self.path = kwargs['path']
            self.writer_path = kwargs['path']
            self.writer = tensorboard.SummaryWriter(self.writer_path)
        else:
            self.path = None
            self.writer = None
            self.writer_path = None
        if "callbacks_param" in kwargs.keys():
            self.callbacks_param = kwargs["callbacks_param"]
        else:
            self.callbacks_param = None

    def reset(self, *args, **kwargs):
        if self.path != None:
            self.writer_path = os.path.join(self.path,str(kwargs['trial'].number)) if 'trial' in kwargs.keys() else self.path
            self.writer = tensorboard.SummaryWriter(self.writer_path)

    def on_epoch_end(self, *args, **kwargs):
        #Write train and validation loss and accuracy
        if self.writer != None:
            epoch = kwargs['epoch']
            history = kwargs['history']
            if self.callbacks_param["training_loss"] or self.callbacks_param["training_loss"] == 'True':
                self.writer.add_scalars('metrics/loss', {'training':history['loss'][epoch]}, epoch)
            if self.callbacks_param["training_accuracy"] or self.callbacks_param["training_accuracy"] == 'True':
                self.writer.add_scalars('metrics/accuracy', {'training':history['accuracy'][epoch]}, epoch)
            if self.callbacks_param["validation_loss"] or self.callbacks_param["validation_loss"] == 'True':
                self.writer.add_scalars('metrics/loss', {'validation':history['val_loss'][epoch]}, epoch)
            if self.callbacks_param["validation_accuracy"] or self.callbacks_param["validation_accuracy"] == 'True':
                self.writer.add_scalars('metrics/accuracy', {'validation':history['val_accuracy'][epoch]}, epoch)

class GuessingEntropyCallback(Callback):
    """Creates a callback that will be called when the epoch end of the training, and will compute the guessing entropy of the model.
    If writer is not None, the guessing entropy will be written in the tensorboard as a plot.
    """
    def __init__(self, traceset, subkey=[0], number_of_key_hypothesis=256, epochs=200, rate=10, n_repeat=5, r=0.6, exp_path="experiments/default/", writer_path=None, *args, **kwargs):
        """ Initializes the callback
        
        :param traceset: Attack dataset
        :type traceset: dict
        :param subkey: List of the indexes of the byte key to target (default [0])
        :type subkey: List[int] or int
        :param number_of_key_hypothesis: Number of key hypothesis to consider (default 256)
        :type number_of_key_hypothesis: int, optional
        :param epochs: Number of epochs to consider (default 200)
        :type epochs: int, optional
        :param rate: Interval of epochs between each run of epoch_end() (default 10)
        :type rate: int, optional
        :param n_repeat: Number of repeat to compute GGE (default 20)
        :type n_repeat: int, optional
        :param r: Rate of the generalized guessing entropy (default 0.6)
        :type r: float, optional
        :param writer: Tensorboard writer
        :type writer: tf.summary.FileWriter, optional
        """
        super(GuessingEntropyCallback, self).__init__(*args, **kwargs)
        self.name = 'gge'
        self.epochs = epochs
        self.subkey = subkey if isinstance(subkey, list) else [subkey]
        self.nk = len(self.subkey)
        self.traceset = {"x_attack":traceset["x_val"].to(device), "m_attack":traceset["m_val"]}
        self.attack_size = len(self.traceset["x_attack"])
        self.repeat = n_repeat
        self.r = r # Percentage of traces used to computes gge
        self.Q = int((1-self.r)*self.attack_size)#number of traces used in GGE
        self.number_of_key_hypothesis = number_of_key_hypothesis
        self.rate = int(rate)
        self.e_path = exp_path
        self.exp_path = exp_path
        if rate != -1:
            os.makedirs(exp_path, exist_ok=True)
        self.ge_index = 0
        self.n_ge = int(self.epochs/self.rate)
        self.guessing_entropy = torch.zeros((self.nk,self.n_ge, self.Q)).to(device)
        self.pred = torch.zeros((self.nk, self.Q, self.number_of_key_hypothesis)).to(device)
        self.last_guessing_entropy = torch.zeros((self.nk, self.Q)).to(device)
        self.w_path = writer_path
        self.writer_path = writer_path
        self.writer = tensorboard.SummaryWriter(self.writer_path) if self.writer_path != None else None
        self.patience = 0
        self.stopping_threshold = 10
        self.fit_value = 0
        self.best_fit_value = number_of_key_hypothesis/2

    def reset(self, *args, **kwargs):
        #Renew parameters and reset state
        #TODO: diff between tensorboard_path and experiment_path
        self.epochs = kwargs['epochs'] if 'epochs' in kwargs.keys() else self.epochs
        self.subkey = kwargs['subkey'] if 'subkey' in kwargs.keys() else self.subkey
        self.nk = len(self.subkey)
        self.repeat = kwargs['n_repeat'] if 'n_repeat' in kwargs.keys() else self.repeat
        self.r = kwargs['r'] if 'r' in kwargs.keys() else self.r
        self.Q = int((1-self.r)*self.attack_size)
        self.rate = kwargs['rate'] if 'rate' in kwargs.keys() else self.rate
        self.n_ge = int(self.epochs/self.rate)
        self.guessing_entropy = torch.zeros((self.nk,self.n_ge, self.Q)).to(device)
        self.pred = torch.zeros((self.nk, self.Q, self.number_of_key_hypothesis)).to(device)
        self.last_guessing_entropy = torch.zeros((self.nk, self.Q)).to(device)
        self.patience = 0
        self.stopping_threshold = 10
        self.fit_value = 0
        self.best_fit_value = self.number_of_key_hypothesis/2
        if self.writer != None:
            self.writer_path = os.path.join(self.w_path,str(kwargs['trial'].number)) if 'trial' in kwargs.keys() else self.w_path
            self.writer = tensorboard.SummaryWriter(self.writer_path)
        if self.e_path != None:
            self.exp_path = os.path.join(self.e_path,str(kwargs['trial'].number)) if 'trial' in kwargs.keys() else self.e_path
            os.makedirs(self.exp_path, exist_ok=True)

    def on_train_begin(self, *args, **kwargs):
        self.ge_index = 0
        self.guessing_entropy.zero_()
        return 0

    def on_epoch_end(self, model, epoch, *args, **kwargs):
        """ Function called on every epoch end.

        :param model: The model to get predicitons from.
        :type model: torch.nn.Module
        :param epoch: Index of the current epoch of the training phase.
        :type epoch: int
        """
        if (epoch+1)%self.rate==0:#trigger every 'rate' epochs
            #make predictions for the validation set
            output_probabilities = model(self.traceset["x_attack"])
            if isinstance(output_probabilities, torch.Tensor):
                output_probabilities = output_probabilities.reshape((1,self.attack_size,-1))
            known_key = torch.ByteTensor([get_known_key(self.traceset["m_attack"], sk, 0) for sk in self.subkey]).to(device)# knwon_key in attack set should be the same for all traces
            
            #Compute GGE
            for _ in range(self.repeat):
                #shuffle validation set
                p = torch.randperm(self.attack_size)
                self.traceset["x_attack"] = self.traceset["x_attack"][p]
                for k in range(len(output_probabilities)):
                    output_probabilities[k] = output_probabilities[k][p]
                self.traceset["m_attack"] = {k:self.traceset["m_attack"][k][p] for k in self.traceset["m_attack"].keys()}
                #compute GE for one iteration
                self.pred.zero_()
                # for sk in range(self.nk):#TODO:Is it possible to skip this loop and work on all subkeys at once?
                for kg in range(self.number_of_key_hypothesis):
                    pw_out = torch.stack([intermediate_value(torch.ShortTensor([kg]*self.Q), get_attack_parameters(self.traceset["m_attack"], sk, slice(0,self.Q))) for sk in range(self.nk)])
                    for sk in range(self.nk):
                        self.pred[sk,:,kg] += output_probabilities[sk][:self.Q][torch.arange(self.Q), pw_out[sk].tolist()]
                for index in range(self.Q-1,0,-1):
                    self.pred[:, index] = torch.sum(self.pred[:,:index],dim=1)
                for sk in range(self.nk):
                    self.guessing_entropy[sk, self.ge_index] += torch.nonzero(torch.argsort(self.pred[sk],dim=1,descending=True) == known_key[sk])[:,1]# correct key rank and accumulate in GE

            self.guessing_entropy[:, self.ge_index] /= self.repeat
            avg_bytes_ge = torch.sum(self.guessing_entropy[:, self.ge_index],dim=0)/self.nk
            logger.info(avg_bytes_ge)
            #Save GE, for tensorboard
            if self.writer != None:
                fig = plt.figure()
                for sk in range(self.nk):
                    plt.plot(self.guessing_entropy[sk, self.ge_index].cpu(), label="target {}".format(self.subkey[sk]))
                plt.ylim((0,self.number_of_key_hypothesis))
                plt.title('Guessing Entropy @ epoch {}'.format(epoch))
                self.writer.add_figure('images/gge', fig, self.ge_index)
                plt.close()
                self.writer.add_scalars('metrics/gge', {'gge':avg_bytes_ge[-1]}, epoch)

            if 'trial' in kwargs.keys():
                #Report intermediate value metric if chosen metric is gge
                self.fit_value = torch.sum(self.guessing_entropy[:,self.ge_index])/(self.nk*len(self.guessing_entropy[0,self.ge_index]))
                kwargs['trial'].set_user_attr("guessingEntropy", self.guessing_entropy[:,self.ge_index].tolist())
                if 'metric' in kwargs.keys():
                    if kwargs['metric'] == 'gge':
                        kwargs['trial'].report(self.fit_value, step=epoch)

                # Early stopping based on average of GE
                if self.fit_value < self.best_fit_value:
                    self.best_fit_value = self.fit_value
                    self.patience = 0
                else:
                    self.patience += 1
                if self.patience >= self.stopping_threshold:
                    logger.info("early stopping based on gge")
                    model.stop_training = True
            self.ge_index += 1

    def on_train_end(self, *args, **kwargs):
        # Save GE to npy file
        guessingEntropy = self.guessing_entropy.detach().cpu().numpy()
        if 'trial' in kwargs.keys():
            filename = os.path.join(self.exp_path, "GE_trail{}.npy".format(kwargs['trial'].number))
        else:
            filename = os.path.join(self.exp_path, 'GE.npy')
        np.save(filename, guessingEntropy)
        self.patience = 0 #Reset patience
        # Return GE as Dict with a numpy array
        return {"GE":guessingEntropy}
        
    def get_fit_value(self):
        return self.fit_value
    
    def get_guessing_entropy(self):
        return self.guessing_entropy.detach().cpu().numpy()
    

def init_callbacks(callbacks:dict, **kwargs):
    """Initialize callbacks from a dictionary of callbacks.
    
    :param callbacks: Dictionary of callbacks
    :type callbacks: dict
    :return: List of callbacks
    :rtype: list
    """
    callbacks_list = []
    for callback_name in callbacks.keys():
        if callback_name.lower() == 'modelcheckpoint':
            callbacks_list.append(ModelCheckpoint(path=kwargs["exp_path"], 
                                                  period=int(callbacks[callback_name]["period"])))
        elif callback_name.lower() == 'tensorboard':
            callbacks_list.append(TensorboardCallback(path=kwargs['writer_path'], callbacks_param=callbacks[callback_name]))
        elif callback_name.lower() == 'guessingentropy':
            callbacks_list.append(GuessingEntropyCallback(traceset=kwargs["traceset"], 
                                                          subkey=kwargs["parameters"]["subkey"], 
                                                          number_of_key_hypothesis=int(kwargs["parameters"]["num_key_hypotheses"]), 
                                                          epochs=int(kwargs["Hparams"]["num_epochs"]) if 'Hparams' in kwargs.keys() else 100, 
                                                          rate=int(callbacks[callback_name]["ge_rate"]), 
                                                          n_repeat=int(callbacks[callback_name]["n_repeat"]), 
                                                          r=float(callbacks[callback_name]["r"]), 
                                                          exp_path=kwargs["exp_path"], 
                                                          writer_path=kwargs["writer_path"]))
        else:
            logger.warning("Callback {} not implemented".format(callback_name))
    return callbacks_list


# eSCAPyDL - a Deep Learning Side-Channel Analysis Python Framework
# Copyright (C) 2023  Weissbart Léo

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import torch

# Sbox
sbox = torch.ByteTensor([
    0x63, 0x7C, 0x77, 0x7B, 0xF2, 0x6B, 0x6F, 0xC5, 0x30, 0x01, 0x67, 0x2B, 0xFE, 0xD7, 0xAB, 0x76,
    0xCA, 0x82, 0xC9, 0x7D, 0xFA, 0x59, 0x47, 0xF0, 0xAD, 0xD4, 0xA2, 0xAF, 0x9C, 0xA4, 0x72, 0xC0,
    0xB7, 0xFD, 0x93, 0x26, 0x36, 0x3F, 0xF7, 0xCC, 0x34, 0xA5, 0xE5, 0xF1, 0x71, 0xD8, 0x31, 0x15,
    0x04, 0xC7, 0x23, 0xC3, 0x18, 0x96, 0x05, 0x9A, 0x07, 0x12, 0x80, 0xE2, 0xEB, 0x27, 0xB2, 0x75,
    0x09, 0x83, 0x2C, 0x1A, 0x1B, 0x6E, 0x5A, 0xA0, 0x52, 0x3B, 0xD6, 0xB3, 0x29, 0xE3, 0x2F, 0x84,
    0x53, 0xD1, 0x00, 0xED, 0x20, 0xFC, 0xB1, 0x5B, 0x6A, 0xCB, 0xBE, 0x39, 0x4A, 0x4C, 0x58, 0xCF,
    0xD0, 0xEF, 0xAA, 0xFB, 0x43, 0x4D, 0x33, 0x85, 0x45, 0xF9, 0x02, 0x7F, 0x50, 0x3C, 0x9F, 0xA8,
    0x51, 0xA3, 0x40, 0x8F, 0x92, 0x9D, 0x38, 0xF5, 0xBC, 0xB6, 0xDA, 0x21, 0x10, 0xFF, 0xF3, 0xD2,
    0xCD, 0x0C, 0x13, 0xEC, 0x5F, 0x97, 0x44, 0x17, 0xC4, 0xA7, 0x7E, 0x3D, 0x64, 0x5D, 0x19, 0x73,
    0x60, 0x81, 0x4F, 0xDC, 0x22, 0x2A, 0x90, 0x88, 0x46, 0xEE, 0xB8, 0x14, 0xDE, 0x5E, 0x0B, 0xDB,
    0xE0, 0x32, 0x3A, 0x0A, 0x49, 0x06, 0x24, 0x5C, 0xC2, 0xD3, 0xAC, 0x62, 0x91, 0x95, 0xE4, 0x79,
    0xE7, 0xC8, 0x37, 0x6D, 0x8D, 0xD5, 0x4E, 0xA9, 0x6C, 0x56, 0xF4, 0xEA, 0x65, 0x7A, 0xAE, 0x08,
    0xBA, 0x78, 0x25, 0x2E, 0x1C, 0xA6, 0xB4, 0xC6, 0xE8, 0xDD, 0x74, 0x1F, 0x4B, 0xBD, 0x8B, 0x8A,
    0x70, 0x3E, 0xB5, 0x66, 0x48, 0x03, 0xF6, 0x0E, 0x61, 0x35, 0x57, 0xB9, 0x86, 0xC1, 0x1D, 0x9E,
    0xE1, 0xF8, 0x98, 0x11, 0x69, 0xD9, 0x8E, 0x94, 0x9B, 0x1E, 0x87, 0xE9, 0xCE, 0x55, 0x28, 0xDF,
    0x8C, 0xA1, 0x89, 0x0D, 0xBF, 0xE6, 0x42, 0x68, 0x41, 0x99, 0x2D, 0x0F, 0xB0, 0x54, 0xBB, 0x16
    ])


# eSCAPyDL - a Deep Learning Side-Channel Analysis Python Framework
# Copyright (C) 2023  Weissbart Léo

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os,sys
os.environ['GDK_BACKEND'] = 'x11'
import datetime
import logging
logging.basicConfig(format='%(message)s')
logger = logging.getLogger()
logger.setLevel(logging.INFO)
import signal
import shlex
import argparse
import time
import subprocess

def main():
    #Parse arguments
    #TODO: remove default dataset and model and default to help
    parser = argparse.ArgumentParser(description='eSCAPyDL: Deep Leaning SCA PyTorch framework. Training models for tracesets with SCA metrics.')
    parser.add_argument('-d', '--dataset', help='The dataset to use for training. (All datasets have different functions for intermediate value and power model).', default="dpav4")
    group = parser.add_mutually_exclusive_group()
    group.add_argument('-t', '--train', help='Train the model', action='store_true')
    group.add_argument('-o', '--optimize', help='Optimize the model', action='store_true')
    parser.add_argument('-m', '--model', help='Type of NN to use (choices: MLP, CNN) CNN is the onetrace implementation', default="MLP")
    parser.add_argument('-s', '--study', help='Optuna study name, for parallelization of study')
    parser.add_argument('-u', '--database', help='Optuna database name', default="sqlite:///cnn_sca.db")
    parser.add_argument('-g', '--gui', help='Start graphical interface', action='store_true', default=False)
    parser.add_argument('-j', '--json', help='Input data from json file')
    parser.add_argument('-a', '--timestamp', help='Timestamp of the experiment', default=time.strftime("%Y%m%d_%H%M%S",time.localtime()))
    parser.add_argument('-v', '--verbose', help='Verbose level', action='count', default=0)
    parser.add_argument('-l', '--log', help='Log into a file at path', default='')
    args = parser.parse_args()
    if args.verbose == 1:
        logger.setLevel(logging.DEBUG)
    if args.log != '':
        file_handler = logging.FileHandler(args.log, mode='w', encoding='utf-8')
        file_handler.setFormatter(logging.Formatter('%(asctime)s %(levelname)s %(message)s', datefmt='%d/%m/%Y %H:%M:%S'))
        logger.addHandler(file_handler)
    if args.train == False and args.optimize == False:
        args.train = True

    logger.debug(args)
    if args.gui:
        command = sys.executable + " {} -l '{}'".format(os.path.join(os.path.dirname(__file__), "gui.py"), args.log)
    else:
        if args.train:
            command = sys.executable + " {} -d {} -m {} -j {} -a {}".format(os.path.join(os.path.dirname(__file__), "train.py"), args.dataset, args.model, args.json, args.timestamp)
        elif args.optimize:
            command = sys.executable + " {} -d {} -m {} -j {} -a {} -s {} -u {}".format(os.path.join(os.path.dirname(__file__),  "optimize.py"), args.dataset, args.model, args.json, args.timestamp, args.study, args.database)
        else:
            logger.error("Error: No action specified. Use '-t' to start straining or '-o' to start optimization")
            exit(1)
    if args.verbose > 0:
        command += " -v"
    logging.debug("Command: {}".format(command))
    #Run command
    p_args = shlex.split(command)
    try:
        start_time = datetime.datetime.now().replace(microsecond=0)
        process = subprocess.Popen(p_args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        logger.debug("Process {} started".format(process.pid))
        while process.poll() is None:
            log_out = process.stdout.readline().decode('utf-8').strip()
            if log_out != '':
                logger.info(log_out)
        logger.debug("Process {} done".format(process.pid))
        stop_time = datetime.datetime.now().replace(microsecond=0)
        if not args.gui:
            logger.info("----------------------\n  The operation was \ncompleted successfully\n Clear time {}\n----------------------".format(stop_time-start_time))
    except KeyboardInterrupt:
        logger.warning("Got Keyboard interrupt")
        logger.debug("Killing process {}".format(process.pid))
        process.send_signal(signal.SIGINT)
        while process.poll() is None:
            log_out = process.stdout.readline().decode('utf-8').strip()
            if log_out != '':
                logger.info(log_out)
    except Exception as e:
        logger.warning(e)
    logger.debug("Program Exited")
    
if __name__ == "__main__":
    sys.exit(main())


import os
import torch
import helpers.aes_helper as aes

#Dataset parameters (dpav4 containing mask encryption leakage of all 16 bytes)
parameters = {
    "dataset_name"        : '/mnt/E_drive/datasets/dpav4/v2/encryption_dpav42_10000.trs',
    "dataset_size"        : 4500,
    "training_size"       : 3500,
    "val_size"            : 1000,
    "subkey"              : 15,
    "num_key_hypotheses"  : 256,
    "n_samples"           : 40000,
    "r"                   : 0.6,
    "n_repeat"            : 5,
    "ge_rate"             : 5,
    "shuffle"             : True,
}

model = {
    "num_epochs"          : 100,
    "loss_function"       : torch.nn.CrossEntropyLoss(),
    "bs"                  : 128,
    "lr"                  : 1e-3,
}

callbacks = {
    "modelcheckpoint":{
        "period":-1,
    },
    "guessingentropy":{
        "ge_rate":5,
        "n_repeat":5,
        "r":0.6,
    },
    "tensorboard":{
        "training_accuracy":True,
        "validation_accuracy":True,
        "training_loss":True,
        "validation_loss":True,
    },
}

#the mask is fixed and public. a random number "offset" is drawn at the beggining (here known here) and shift the mask.
#AES RSM schedule is : x = plaintext ^ mask_offset and then add the round key and do the masked subbytes: sbox[x ^ key ^ M_i] ^ M_i+1, where i is offset+subkey+round (here round is 0)
mask = torch.ByteTensor([
    0x03, 0x0c, 0x35, 0x3a, 0x50, 0x5f, 0x66, 0x69, 0x96, 0x99, 0xa0, 0xaf, 0xc5, 0xca, 0xf3, 0xfc
])

def get_attack_parameters(traceset, subkey, index):
    """Returns a dictionary of parameters to be used to search for attack .
    To be used before calling intermediate_value().

    :param traceset: Traceset to use
    :type traceset: dict
    :param subkey: Subkey to attack
    :type subkey: int
    :param index: Index of the traceset to use
    :type index: int or slice
    :return: Dictionary of parameters
    :rtype: dict
    """        
    return {'p':traceset["plaintext"][index,subkey], 
            'k':traceset["key"][index,subkey], 
            'o':traceset["offset"][index,subkey], 
            'subkey':subkey}

def get_known_key(traceset, subkey, index):
    return traceset['key'][index][subkey]

def intermediate_value(keyguess, trace_parameters):
    """Return the intermediate value of a key in the SBox . 
    First xor corresponds to the plaintext blinding, then we do addRoundKey with key guess and apply rsm fo subBytes

    :param trace_parameters: Dictionary containing the trace parameters for intermediate value computation
    :type trace_parameters: dict
    :return: Intermediate value
    :rtype: Any
    """        
    if isinstance(trace_parameters['p'], torch.Tensor):
        return torch.index_select(aes.sbox, 0, (trace_parameters['p'] ^ keyguess).int()) ^ torch.index_select(mask, 0, ((trace_parameters['o']+1)%16).int())
    else:
        return aes.sbox.numpy()[(trace_parameters['p'] ^ keyguess)] ^ mask.numpy()[(trace_parameters['o']+1)%16]



# eSCAPyDL - a Deep Learning Side-Channel Analysis Python Framework
# Copyright (C) 2023  Weissbart Léo

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sys, logging
import numpy as np
from itertools import chain
import torch

reg_size = 64
ASCON_128_IV_int = 0x80400c0600000000
ASCON_128_IV = (0x80400c0600000000).to_bytes(16, 'big')

ASCON_SBOX_NP = np.array([0x04, 0x0b, 0x1f, 0x14, 0x1a, 0x15, 0x09, 0x02, 
              0x1b, 0x05, 0x08, 0x12, 0x1d, 0x03, 0x06, 0x1c, 
              0x1e, 0x13, 0x07, 0x0e, 0x00, 0x0d, 0x11, 0x18, 
              0x10, 0x0c, 0x01, 0x19, 0x16, 0x0a, 0x0f, 0x17])
ASCON_SBOX_NP_DIM = [np.array([ASCON_SBOX_NP[x] for x in range(1<<4)]), np.array([ASCON_SBOX_NP[x] for x in range(1<<4, 1<<5)])]#If first bit is 0 or 1
ASCON_SBOX = torch.ByteTensor(ASCON_SBOX_NP)

#diffusion layer on register x_i
dr = [[45,36],
      [3,25],
      [63,58],
      [54,47],
      [57,23]]

def bit_value(i, register_value):
    return (register_value>>(reg_size-1-i))&1
    
def select_bit_bytes(i, array):
    """Returns the bit index `i` from the array of bytes `array`
    If array have 2 dimensions, it will select bits given the first axis.
    
    :param i: index of the bit to select.
    :type i: int
    :param aray: array of bytes in bytes/int format.
    :type array: list
    """
    if isinstance(array,np.ndarray) or isinstance(array, torch.Tensor):
        if len(array.shape)==1:
            i = i % (len(array)*8)
            return (array[i//8]>>(7-(i%8)))&1
        elif len(array.shape)==2:
            i = i % (len(array[0])*8)
            return (array[:,i//8]>>(7-(i%8)))&1
    elif isinstance(array, bytes) or isinstance(array, list):
        i = i % (len(array)*8)
        return (array[i//8]>>(7-(i%8)))&1
    else: 
        logging.error("select_bit_bytes: array type not supported, got {}".format(type(array)))
        logging.exception('')

#-----------------------------------------#
# Intermediate value of SBox and Linear Diffusion Layer
reg_num_0 = 4
reg_num_1 = 1
def intermediate_value(keyguess, trace_parameters):
    """Return the intermediate value of a key in the SBox . 
    nonce0 and nonce1 should be byte arrays of size (8) as they both are half of the nonce length (16 bytes).
    k should be a 3-bit value as an integer. This bit corresponds to the part of the key concerned in the attack given subkey.

    :param trace_parameters: Dictionary containing the trace parameters for intermediate value computation
    :type trace_parameters: dict
    :return: Intermediate value
    :rtype: Any
    """        
    if trace_parameters['subkey']<reg_size:
        y0_i = lambda x0,x1,x3,x4: x4&(x1^1)^x3
        # y0_i = lambda x0,x1,x3,x4: x1&(x0+x4+1)+x3+x4
        res = (y0_i(select_bit_bytes(trace_parameters['subkey']%reg_size           ,ASCON_128_IV),   (keyguess>>2)&1, select_bit_bytes(trace_parameters['subkey']%reg_size           , trace_parameters['nonce0']), select_bit_bytes(trace_parameters['subkey']%reg_size           , trace_parameters['nonce1']))) \
            ^ (y0_i(select_bit_bytes((trace_parameters['subkey']+dr[reg_num_0][0])%reg_size,ASCON_128_IV),   (keyguess>>1)&1, select_bit_bytes((trace_parameters['subkey']+dr[reg_num_0][0])%reg_size, trace_parameters['nonce0']), select_bit_bytes((trace_parameters['subkey']+dr[4][0])%reg_size, trace_parameters['nonce1']))) \
            ^ (y0_i(select_bit_bytes((trace_parameters['subkey']+dr[reg_num_0][1])%reg_size,ASCON_128_IV),   (keyguess>>0)&1, select_bit_bytes((trace_parameters['subkey']+dr[reg_num_0][1])%reg_size, trace_parameters['nonce0']), select_bit_bytes((trace_parameters['subkey']+dr[4][1])%reg_size, trace_parameters['nonce1'])))
        return res
    elif trace_parameters['subkey']<2*reg_size:
        y0_i = lambda x0,x12,x3,x4: x3&(x12^1)^x4
        # y0_i = lambda x0,x12,x3,x4: x4&(x3^1)^x12^1
        res = (y0_i(select_bit_bytes(trace_parameters['subkey']%reg_size           ,ASCON_128_IV),(keyguess>>2)&1, select_bit_bytes(trace_parameters['subkey']%reg_size           , trace_parameters['nonce0']), select_bit_bytes(trace_parameters['subkey']%reg_size           , trace_parameters['nonce1']))) \
            ^ (y0_i(select_bit_bytes((trace_parameters['subkey']+dr[reg_num_0][0])%reg_size,ASCON_128_IV),(keyguess>>1)&1, select_bit_bytes((trace_parameters['subkey']+dr[reg_num_1][0])%reg_size, trace_parameters['nonce0']), select_bit_bytes((trace_parameters['subkey']+dr[reg_num_1][0])%reg_size, trace_parameters['nonce1']))) \
            ^ (y0_i(select_bit_bytes((trace_parameters['subkey']+dr[reg_num_0][1])%reg_size,ASCON_128_IV),(keyguess>>0)&1, select_bit_bytes((trace_parameters['subkey']+dr[reg_num_1][1])%reg_size, trace_parameters['nonce0']), select_bit_bytes((trace_parameters['subkey']+dr[reg_num_1][1])%reg_size, trace_parameters['nonce1'])))
        return res
    
def select_subkey(i, key, array=False):
    if i<reg_size:
        sub_key = [select_bit_bytes((i)%reg_size         , key[0:8]), 
                   select_bit_bytes((i+dr[reg_num_0][0])%reg_size, key[0:8]), 
                   select_bit_bytes((i+dr[reg_num_0][1])%reg_size, key[0:8])]
    elif i<2*reg_size:
        sub_key = [x^y for x,y in zip([select_bit_bytes((i)%reg_size         , key[0:8]), 
                                       select_bit_bytes((i+dr[reg_num_1][0])%reg_size, key[0:8]), 
                                       select_bit_bytes((i+dr[reg_num_1][1])%reg_size, key[0:8])], 
                                      [select_bit_bytes((i)%reg_size         , key[8:16]), 
                                       select_bit_bytes((i+dr[reg_num_1][0])%reg_size, key[8:16]), 
                                       select_bit_bytes((i+dr[reg_num_1][1])%reg_size, key[8:16])]
                                    )
        ]
    if not array:
        sub_key = sub_key[0]<<2 | sub_key[1]<<1 | sub_key[2]
    return sub_key
#-----------------------------------------#

#-----------------------------------------#
# Intermediate value for a chunk of bits after the SBox and Linear Diffusion Layer
def intermediate_value_chunk(keyguess, trace_parameters):
    iv = 0
    chunk_size = 3 # Size in bits of the intermidate_value
    targetByte = trace_parameters['subkey'] # 'subkey' is index of the byte to target
    keyguess_bits = [(keyguess>>i)&1 for i in range(3*chunk_size)]
    for bitNum in range(len(keyguess_bits)//chunk_size):
        trace_parameters['subkey'] = (targetByte*chunk_size) + bitNum # Transform into the bit to target
        sub_keyguess = keyguess_bits[bitNum::chunk_size]
        _ret = 0
        for i, sk in enumerate(sub_keyguess):
            _ret ^= sk<<((len(sub_keyguess)-1-i))
        sub_keyguess = _ret
        iv ^= (intermediate_value(sub_keyguess, trace_parameters)<<((chunk_size-1)-bitNum))
    return iv

def select_subkey_chunk(i, key, array=False):
    chunk_size = 3
    sub_key = []
    if i<reg_size:
        sub_key.append([select_bit_bytes((i+c)%reg_size           , key[0:8]) for c in range(chunk_size)])
        sub_key.append([select_bit_bytes((i+c+dr[4][0])%reg_size  , key[0:8]) for c in range(chunk_size)])
        sub_key.append([select_bit_bytes((i+c+dr[4][1])%reg_size  , key[0:8]) for c in range(chunk_size)])
    elif i<2*reg_size:
        sub_key.append([select_bit_bytes((i+c)%reg_size, key[0:8])          ^ select_bit_bytes((i+c)%reg_size, key[8:16])          for c in range(chunk_size)])
        sub_key.append([select_bit_bytes((i+c+dr[1][0])%reg_size, key[0:8]) ^ select_bit_bytes((i+c+dr[1][0])%reg_size, key[8:16]) for c in range(chunk_size)])
        sub_key.append([select_bit_bytes((i+c+dr[1][1])%reg_size, key[0:8]) ^ select_bit_bytes((i+c+dr[1][1])%reg_size, key[8:16]) for c in range(chunk_size)])
    sub_key = list(chain.from_iterable(sub_key)) # flatten the array
    if array==False:
        _ret = 0
        for i,e in enumerate(sub_key):
            _ret += e<<((len(sub_key)-1-i))
        sub_key = _ret
    return sub_key
#-----------------------------------------#

#-----------------------------------------#
# SBOX intermediate value. keygess is a 2bit value
def intermediate_value_sbox(keyguess, trace_parameters):
    if 'nonce0' in trace_parameters.keys():
        if isinstance(trace_parameters['nonce0'], torch.Tensor):
            ind = ((select_bit_bytes(trace_parameters['subkey'], ASCON_128_IV))<<4 | 
                    ((keyguess>>1)&1)<<3 | 
                    ((keyguess>>0)&1)<<2 | 
                    (select_bit_bytes(trace_parameters['subkey'], trace_parameters['nonce0'].numpy()))<<1 | 
                    (select_bit_bytes(trace_parameters['subkey'], trace_parameters['nonce1'].numpy()))<<0)
            return torch.index_select(ASCON_SBOX, 0, ind.int())
        else:
            ind = ((select_bit_bytes(trace_parameters['subkey'], ASCON_128_IV))<<4 | 
                    ((keyguess>>1)&1)<<3 | 
                    ((keyguess>>0)&1)<<2 | 
                    (select_bit_bytes(trace_parameters['subkey'], trace_parameters['nonce0']))<<1 | 
                    (select_bit_bytes(trace_parameters['subkey'], trace_parameters['nonce1']))<<0)
            return ASCON_SBOX_NP[ind]
    elif 'nonce' in trace_parameters.keys():
        if isinstance(trace_parameters['nonce'], torch.Tensor):
            ind = torch.IntTensor([((select_bit_bytes(trace_parameters['subkey'], ASCON_128_IV))<<4 | 
                    ((keyguess[id]>>1)&1)<<3 | 
                    ((keyguess[id]>>0)&1)<<2 | 
                    (select_bit_bytes(trace_parameters['subkey'], trace_parameters['nonce'][id][:8].numpy()))<<1 | 
                    (select_bit_bytes(trace_parameters['subkey'], trace_parameters['nonce'][id][8:16].numpy()))<<0) for id in range(len(keyguess))])
            return torch.index_select(ASCON_SBOX, 0, ind.int())
        else:
            ind = ((select_bit_bytes(trace_parameters['subkey'], ASCON_128_IV))<<4 | 
                    ((keyguess>>1)&1)<<3 | 
                    ((keyguess>>0)&1)<<2 | 
                    (select_bit_bytes(trace_parameters['subkey'], trace_parameters['nonce'][:8]))<<1 | 
                    (select_bit_bytes(trace_parameters['subkey'], trace_parameters['nonce'][8:16]))<<0)
            return ASCON_SBOX_NP[ind]
    
def select_subkey_sbox(i, key):
    return (select_bit_bytes(i%(2*reg_size), key))<<1 | select_bit_bytes((i+reg_size)%(2*reg_size), key)
    
#-----------------------------------------#

#-----------------------------------------#
# Chunk SBOX intermediate value. keygess is a chunk_size-bit value for the two key registers
def intermediate_value_sbox_chunk(keyguess, trace_parameters, chunk_size=8):
    if trace_parameters['subkey']*chunk_size<reg_size:
        y0_i = lambda x0,x1,x3,x4: x1&(x0^x4^1)^x3^x4
        res = 0
        for i in range(chunk_size):
            res |= (y0_i(select_bit_bytes((trace_parameters['subkey']*chunk_size+i)%reg_size,ASCON_128_IV), 
                         (keyguess>>(chunk_size-1 - i))&1, 
                         select_bit_bytes((trace_parameters['subkey']*chunk_size+i)%reg_size, trace_parameters['nonce0']), 
                         select_bit_bytes((trace_parameters['subkey']*chunk_size+i)%reg_size, trace_parameters['nonce1']))) << (chunk_size-1 - i)
        return res
    elif trace_parameters['subkey']*chunk_size<2*reg_size:
        y0_i = lambda x12,x3,x4: x4&(x3^1)^x12^1
        res = 0
        for i in range(chunk_size):
            res |= (y0_i((keyguess>>(chunk_size-1 - i))&1, 
                         select_bit_bytes((trace_parameters['subkey']*chunk_size+i)%reg_size, trace_parameters['nonce0']), 
                         select_bit_bytes((trace_parameters['subkey']*chunk_size+i)%reg_size, trace_parameters['nonce1']))) << (chunk_size-1 - i)
        return res
    
def select_subkey_sbox_chunk(i, key, chunk_size=8):
    ret = 0
    for bit in range(chunk_size):
        ret |= select_bit_bytes(i*chunk_size + bit, key)<<(chunk_size-1 - bit)
    if i*chunk_size>=reg_size:
        for bit in range(chunk_size):
            ret ^= select_bit_bytes((i*chunk_size + bit)%64, key)<<(chunk_size-1 - bit)
    return ret

#-----------------------------------------#